# -*- coding: utf-8 -*-

import numpy as np
from matplotlib import pyplot as pp
from mpl_toolkits.mplot3d import Axes3D
import interp
from scipy import ndimage as ndim
from scipy import optimize as opt
from scipy.signal import convolve, convolve2d
import hpfilter

#%% Poglavje 1

def imLoadRaw2d(filename, width, height, dtype=np.uint8, order='xy'):
  '''
  Funkcija naloži v pomnilniki dvorazsežno sliko zapiano v surov binarnem
  formatu. Podatkovno polje naložene slike je velikosti (height, width).

  Parametri
  ---------
  filename: str
    Ime datoteke, v kateri je shranjena slika.
  width: int
    Širina slike izražena v slikovnih elementih (vzdolž x koordinatne osi).
  height: int
    Višina slike izražena v slikovnih elementih (vzdolž y koordinatne osi).
  dtype: numpy.dtype
    Podatkovni tip slikovnega elementa.
  order : str, neobvezno
    Vrstni red zapisa slikovnih elementov v spominu 'xy' ali 'yx'.

    Uporabi 'xy' - za sliko shranjeno po vrsticah.

    Uporabi 'yx' - za sliko shranjeno po stolpcih

  Vrne
  ----
  img: numpy.ndarray
    Podatkovno polje velikosti (height, width).

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> slika = imLoadRaw2d('./poglavje_1/rezultati/mr_217x181_uint8.raw', 217, 181)
  >>>
  >>> pp.figure()
  >>> pp.imshow(slika, cmap='gray')
  >>> pp.title('Slika velikosti {} in tipa {}'.format(str(slika.shape), str(slika.dtype)))
  '''
  order = str(order).lower()
  data = np.fromfile(filename, dtype=dtype)  
  if order == 'xy':
    data.shape = (height, width)
  elif order == 'yx':
    data.shape = (width, height)
    data = data.transpose()
  else:
    raise ValueError(
      'Vrednost vhodnega parametra "order" je lahko "xy" ali "yx"!')

  return data

def imSaveRaw2d(filename, data):
  '''
  Shrani dvorazsežno sliko v surovo binarno datoteko. 
  Podatkovni tip in vrstni red zapisa določa podatkovno polje.

  Parametri
  ---------
  filename: str
    Ime datoteke, v katero bo shranjena slika.
  data: numpy.ndarray
    Podatkovno polje slike.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> slika = np.tile(np.arange(256, dtype=np.uint8), [128, 1]) # (128, 256)
  >>> imSaveRaw2d('./poglavje_1/rezultati/gradient_256x128_uint8.raw', slika)
  >>>
  >>> slika = imLoadRaw2d('./poglavje_1/rezultati/gradient_256x128_uint8.raw', 256, 128)
  >>>
  >>> pp.figure()
  >>> pp.imshow(slika, cmap='gray')
  >>> pp.title('Slika velikosti {} in tipa {}'.format(str(slika.shape), str(slika.dtype)))
  '''
  data.tofile(filename)

    
def imLoadRaw3d(filename, width, height, depth, dtype=np.uint8, order='xyz'):
  '''
  Funkcija naloži v pomnilniki trirazsežno sliko zapisano v surovem binarnem
  formatu. Podatkovno polje naložene slike je velikosti (depth, height, width).

  Parametri
  ---------
  filename: str
    Ime datoteke, v kateri je shranjena slika.
  width: int
    Širina slike izražena v slikovnih elementih (vzdolž x koordinatne osi).
  height: int
    Višina slike izražena v slikovnih elementih (vzdolž y koordinatne osi).
  depth: int
    Število rezin/slikovnih elementov v tretji osi (vzdolž z koordinatne osi).
  dtype: numpy.ndarray
    Podatkovni tip slikovnega elementa.
  order : str
    Vrstni red zapisa slikovnih elementov v spominu je lahko eden
    izmed 'xyz', 'xzy', 'yxz', 'yzx', 'zxy', 'zyx'.

    Uporabi 'xyz' - za sliko shranjeno po vrsticah.

    ...

    Uporabi 'zyx' - za sliko shranjeno po globini

  Vrne
  ----
  img: numpy.ndarray
    Podatkovno polje velikosti (depth, height, width).

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> slika = imLoadRaw3d('./poglavje_1/mr_217x181x181_uint8.raw', 217, 181, 181)
  >>>
  >>> pp.figure()
  >>> pp.imshow(slika[90], cmap='gray')
  >>> pp.title('Prerez [90, :, :] slike velikosti {} in tipa {}'.format(str(slika.shape), str(slika.dtype)))
  '''
  order = str(order).lower()
  data = np.fromfile(filename, dtype=dtype)
  if order == 'xyz':
    data.shape = (depth, height, width)
  elif order == 'xzy':
    data.shape = (height, depth, width)
    data = data.transpose((1,0,2))
  elif order == 'yxz':
    data.shape = (depth, width, height)
    data = data.transpose((0,2,1))
  elif order == 'yzx':
    data.shape = (width, depth, height)
    data = data.transpose((1,2,0))
  elif order == 'zxy':
    data.shape = (height, width, depth)
    data = data.transpose((2,0,1))
  elif order == 'zyx':
    data.shape = (width, height, depth)
    data = data.transpose((2,1,0))
  else:
    raise ValueError('Vrednost vhodnega parametra "order" je lahko ' \
                     ' "xyz", "xzy", "yxz", "yzx", "zxy", "zyx"!')
  return data

def imSaveRaw3d(filename, data):
  '''
  Shrani trirazsežno sliko v surovo binarno datoteko.
  Podatkovni tip in vrstni red zapisa določa podatkovno polje.

  Parametri
  ---------
  filename: str
    Ime datoteke, v katero bo shranjena slika.
  data: numpy.ndarray
    Podatkovno polje slike.

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> slika = imLoadRaw3d('./poglavje_1/mr_217x181x181_uint8.raw', 217, 181, 181)
  >>> imSaveRaw3d('./tmp/mr_100x181x181_uint8.raw', slika[:, :, :100])
  >>>
  >>> slika = imLoadRaw3d('./tmp/mr_100x181x181_uint8.raw', 100, 181, 181)
  >>>
  >>> pp.imshow(slika[50], cmap='gray')
  >>> pp.title('Prerez [50, :, :] slike velikosti {} in tipa {}'.format(str(slika.shape), str(slika.dtype)))
  '''
  data.tofile(filename)

def imGrid2d(width, height, dx=1, dy=1, xoffset=0, yoffset=0):
  '''
  Ustvari koordinatno mrežo točk 2d slike, ki pripada podatkovnemu polju
  velikosti (depth, height, width).

  Parametri
  ---------
  width: int
    Širina slike izražena v slikovnih elementih (v smeri x koordinatne osi).
  height: int
    Višina slike izražena v slikovnih elementih (v smeri y koordinatne osi).
  dx: float
    Velikost slikovnega elemnta v smeri x koordinatne osi.
  dy: float
    Velikost slikovnega elemnta v smeri y koordinatne osi.
  xoffset: float
    X koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.
  yoffset: float
    Y koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.

  Vrne
  ----
  Y: numpy.ndarray
    Podatkovno polje Y koordinatnih točk slikovnih elementov 2d slike.
  X: numpy.ndarray
    Podatkovno polje X koordinatnih točk slikovnih elementov 2d slike.

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> Y, X = im_grid_2d(10, 5)
  >>> Yc, Xc = im_grid_2d(10, 5, xoffset='center', yoffset='center')
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 3, 1)
  >>> pp.title('Brez zamika')
  >>> pp.plot(X, Y, 'xr')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> pp.title('Koordinatno izhodišče v težišču')
  >>> pp.plot(Xc, Yc, 'og')
  >>>
  >>> pp.subplot(1, 3, 3)
  >>> pp.title('Združeno')
  >>> pp.plot(X, Y, 'xr')
  >>> pp.plot(Xc, Yc, 'og')
  '''
  if yoffset == 'center':
    yoffset = -0.5*(height - 1)*dy
  if xoffset == 'center':
    xoffset = -0.5*(width - 1)*dx

  x = np.arange(xoffset, xoffset + width*dx, dx, dtype=np.float)
  y = np.arange(yoffset, yoffset + height*dy, dy, dtype=np.float)
  return np.meshgrid(y, x, indexing='ij')

def imGrid3d(width, height, depth, dx=1, dy=1, dz=1,
             xoffset=0, yoffset=0, zoffset=0):
  '''
  Ustvari koordinatno mrežo točk 3d slike, ki pripada podatkovnemu polju
  velikosti (depth, height, width).

  Parametri
  ---------
  width: int
    Širina slike izražena v slikovnih elementih (v smeri x koordinatne osi).
  height: int
    Višina slike izražena v slikovnih elementih (v smeri y koordinatne osi).
  depth: int
    Višina slike izražena v slikovnih elementih (v smeri z koordinatne osi).
  dx: float
    Velikost slikovnega elemnta v smeri x koordinatne osi.
  dy: float
    Velikost slikovnega elemnta v smeri y koordinatne osi.
  dz: float
    Velikost slikovnega elemnta v smeri z koordinatne osi.
  xoffset: float
    X koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.
  yoffset: float
    Y koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.
  zoffset: float
    Z koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.

  Vrne
  ----
  Z: numpy.ndarray
    Podatkovno polje Z koordinatnih točk slikovnih elementov 3d slike.
  Y: numpy.ndarray
    Podatkovno polje Y koordinatnih točk slikovnih elementov 3d slike.
  X: numpy.ndarray
    Podatkovno polje X koordinatnih točk slikovnih elementov 3d slike.

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>> from mpl_toolkits.mplot3d import Axes3D
  >>>
  >>> Z, Y, X = im_grid_3d(20, 10, 5)
  >>> Zc, Yc, Xc = im_grid_3d(20, 10, 5, xoffset='center', yoffset='center', zoffset='center')
  >>>
  >>> fig = pp.figure()
  >>>
  >>> ax = fig.add_subplot(131, projection='3d')
  >>> ax.plot(X.flatten(), Y.flatten(), Z.flatten(), 'xr')
  >>> ax.set_title('Brez zamika')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> ax = fig.add_subplot(132, projection='3d')
  >>> ax.plot(Xc.flatten(), Yc.flatten(), Zc.flatten(), 'og')
  >>> ax.set_title('Koordinatno izhodišče v težišču')
  >>>
  >>> pp.subplot(1, 3, 3)
  >>> ax = fig.add_subplot(133, projection='3d')
  >>> ax.plot(X.flatten(), Y.flatten(), Z.flatten(), 'xr')
  >>> ax.plot(Xc.flatten(), Yc.flatten(), Zc.flatten(), 'og')
  >>> ax.set_title('Združeno')
  '''

  if zoffset == 'center':
    zoffset = -0.5*(depth - 1)*dz
  if yoffset == 'center':
    yoffset = -0.5*(height - 1)*dy
  if xoffset == 'center':
    xoffset = -0.5*(width - 1)*dx

  x = np.arange(xoffset, xoffset + width*dx, dx, dtype=np.float)
  y = np.arange(yoffset, yoffset + height*dy, dy, dtype=np.float)
  z = np.arange(zoffset, zoffset + depth*dz, dz, dtype=np.float)

  return np.meshgrid(z, y, x, indexing='ij')

#%% Poglavje 2

def interp1d(x, xp, fp, method='linear'):
  '''
  Interpolacija 1d diskretnega signala, ki je definiran na ekvidistantnem
  zaporedju monotono naraščajočih točk.
  Funkcijske vrednosti izven definicijskega območja postavimo na 0.

  Parametri
  ---------
  x: numpy.ndarray
      X-koordinate točk, v katerih interpoliramo funkcijske vrednosti fp.
  xp: numpy.ndarray
      Ekvidistantne monotono naraščajoče x koordinate podatkovnih točk.
  fp: numpy.ndarray
      Funkcijske vrednosti v podatkovnih točkah xp.
  method: str,
      Način interpolacije 'linear' (privzeto), ali 'nearest'.

  Vrne
  ----
  fx: numpy.ndarray
    Interpolirane funkcijske vrednosti v točkah f[x].

  Primeri uporabe
  ------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> xp = np.linspace(0, np.pi, 10)
  >>> fp = np.sin(xp)
  >>>
  >>> x = np.linspace(0, np.pi, 30)
  >>>
  >>> f_nearest = interp1d(x, xp, fp, 'nearest')
  >>> f_linear = interp1d(x, xp, fp, 'linear')
  >>>
  >>> pp.figure()
  >>> pp.plot(xp, fp, '-xg', label='Podatki')
  >>> pp.plot(x, f_linear, '-or', label='Linearna interp.')
  >>> pp.plot(x, f_nearest, '-*k', label='Interp. z najbližjim sosedom')
  >>> pp.legend(loc='lower center')
  '''
  x = np.asarray(x, np.float)
  xp = np.asarray(xp, np.float)
  fp = np.asarray(fp, np.float)

  dx = xp[1] - xp[0]
  Nxp = xp.size
  Nx = x.size
  f = np.zeros_like(x)

  if method == 'linear':
    # v for zanki
    '''
    xshape = x.shape
    x = x.flatten()
    for i in range(Nx):
      fp_ind = (x[i] - xp[0])/dx
      b_dx = fp_ind - np.floor(fp_ind)
      left = np.floor(fp_ind).astype(np.int)
      if fp_ind >= 0 and fp_ind <= Nxp - 1:
          f[i] = (1.0 - b_dx)*fp[left] + b_dx*fp[min(left + 1, Nxp - 1)]
    f.shape = xshape
    '''
    # brez for zanke
    fp_ind = (x - xp[0])/dx
    b_dx = fp_ind - np.floor(fp_ind)
    left = np.floor(fp_ind).astype(np.int)
    left = np.clip(left, 0, Nxp - 1)
    right = np.clip(left + 1, 0, Nxp - 1)
    f = (1.0 - b_dx)*fp[left] + b_dx*fp[right]
    valid = np.logical_and(fp_ind >= 0, fp_ind <= Nxp - 1)
    f[np.logical_not(valid)] = 0.0

  elif method == 'nearest':
    # v for zanki
    '''
    f = np.zeros_like(x)
    for i in range(Nx):
      indf = np.round((x[i] - xp[0])/dx)
      ind = indf.astype(np.int)
      if ind >= 0 or ind <= Nxp:
          f[i] = fp[ind]
    '''
    # brez for zanke
    fp_ind = (x - xp[0])/dx
    ind = np.round(fp_ind).astype(np.int)
    valid = np.logical_and(ind >=0, ind <= Nxp)
    f[valid] = fp[ind[valid]]

  else:
    raise ValueError(
      'Vrednost parametra "method" je lahko '
      '"linear" ali "nearest"!')

  return f

def interp2d(x, y, xp, yp, fp, method='linear'):
  '''
  Interpolacija 2d slike. Izvedba funkcije temelji na uporabi for zank
  in je temu primerno počasna. Za hitro izvedbo interpolacije glej
  funkcijo map_coordinates modula scipy.ndimage ali funkcije modula interp.
  Funkcijske vrednosti izven definicijskega območja postavimo na 0.

  Parametri:
  ----------
  x: numpy.ndarray
    x-koordinate točk, v katerih interpoliramo funkcijske vrednosti fp.
  y: numpy.ndarray
    y-koordinate točk, v katerih interpoliramo funkcijske vrednosti fp.
  xp: numpy.ndarray
    Vektor ekvidistantnih monotono naraščajočih x koordinat podatkovnih točk.
  yp: numpy.ndarray
    Vektor ekvidistantnih monotono naraščajočih y koordinat podatkovnih točk.
  fp: numpy.ndarray
    Funkcijske vrednosti v točkah, ki jih določata vektorja xp in yp.
    Velikost podatkovnega polja fp mora biti enaka (yp.size, xp.size).
  method: str
    Način interpolacije 'linear' (privzeto), ali 'nearest'.

  Vrne:
  -----
  fxy: numpy.ndarray
    Interpolirane funkcijske vrednosti v točkah f[y, x].

  Primeri uporabe
  ------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> xp = np.linspace(-1.0, 1.0, 20)
  >>> yp = np.linspace(-1.0, 1.0, 10)
  >>> Yp, Xp = np.meshgrid(yp, xp, indexing='ij')
  >>> fp = Xp**2 + Yp**2
  >>>
  >>> x = np.linspace(-1.0, 1.0, 60)
  >>> y = np.linspace(-1.0, 1.0, 30)
  >>> Y, X = np.meshgrid(y, x, indexing='ij')
  >>>
  >>> f_linear = interp2d(X, Y, xp, yp, fp)
  >>> f_nearest = interp2d(X, Y, xp, yp, fp, 'nearest')
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 3, 1)
  >>> pp.title('Podatki')
  >>> pp.imshow(fp, cmap='gray')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> pp.imshow(f_nearest, cmap='gray')
  >>> pp.title('Interpolacija z najbližjim osedom')
  >>>
  >>> pp.subplot(1, 3, 3)
  >>> pp.imshow(f_linear, cmap='gray')
  >>> pp.title('Linearn interpolacija')
  '''
  x = np.asarray(x, np.float)
  y = np.asarray(y, np.float)
  xp = np.asarray(xp, np.float)
  yp = np.asarray(yp, np.float)
  fp = np.asarray(fp, np.float)

  xshape = x.shape
  x, y = x.flatten(), y.flatten()

  dx = xp[1] - xp[0]
  dy = yp[1] - yp[0]

  Nx, Ny = x.size, y.size
  Nxp, Nyp = xp.size, yp.size
  if Nx != Ny:
    raise ValueError(
      'Število elementov v x in y mora biti enako!')

  f = np.zeros([Nx])
  if method == 'linear':
    # v for zanki
    '''
    for i in range(Nx):
      fp_indx = (x[i] - xp[0])/dx
      fp_indy = (y[i] - yp[0])/dy
      w_dx = fp_indx - np.floor(fp_indx)
      h_dy = fp_indy - np.floor(fp_indy)
      indx = np.clip(fp_indx, 0, Nxp - 1)
      indx_plus_1 = np.clip(indx + 1, 0, Nxp - 1)
      indy = np.clip(fp_indy, 0, Nyp - 1)
      indy_plus_1 = np.clip(indy + 1, 0, Nyp - 1)

      a = (1.0 - w_dx)*(1.0 - h_dy)
      b = w_dx*(1.0 - h_dy)
      c = (1.0 - w_dx)*h_dy
      d = w_dx*h_dy
      if fp_indx >= 0 and fp_indx <= Nxp - 1 and \
          fp_indy >= 0 and fp_indy <= Nyp - 1:
        f[i] = a*fp[indy, indx] + \
               b*fp[indy, indx_plus_1] + \
               c*fp[indy_plus_1, indx] + \
               d*fp[indy_plus_1, indx_plus_1]
    '''

    # brez for zanke
    fp_indx = (x - xp[0])/dx
    fp_indy = (y - yp[0])/dy
    w_dx = fp_indx - np.floor(fp_indx)
    h_dy = fp_indy - np.floor(fp_indy)
    indx = np.clip(fp_indx.astype(np.int), 0, Nxp - 1)
    indx_plus_1 = np.clip(indx + 1, 0, Nxp - 1)
    indy = np.clip(fp_indy.astype(np.int), 0, Nyp - 1)
    indy_plus_1 = np.clip(indy + 1, 0, Nyp - 1)

    a = (1.0 - w_dx)*(1.0 - h_dy)
    b = w_dx*(1.0 - h_dy)
    c = (1.0 - w_dx)*h_dy
    d = w_dx*h_dy
    f = a*fp[indy, indx] + \
        b*fp[indy, indx_plus_1] + \
        c*fp[indy_plus_1, indx] + \
        d*fp[indy_plus_1, indx_plus_1]
    valid_x = np.logical_and(fp_indx >= 0, fp_indx <= Nxp - 1)
    valid_y = np.logical_and(fp_indy >= 0, fp_indy <= Nyp - 1)
    valid = np.logical_and(valid_x, valid_y)
    f[np.logical_not(valid)] = 0

  elif method == 'nearest':
    # v for zanki
    '''
    for i in range(Nx):
      fp_indx = np.round((x[i] - xp[0])/dx)
      fp_indy = np.round((y[i] - yp[0])/dy)
      indx = np.clip(fp_indx, 0, Nxp - 1)
      indy = np.clip(fp_indy, 0, Nyp - 1)
      if fp_indx >= 0 and fp_indx <= Nxp - 1 and \
          fp_indy >= 0 and fp_indy <= Nyp - 1:
        f[i] = fp[indy, indx] 
    '''

    # brez for zanke
    fp_indx = np.round((x - xp[0])/dx)
    fp_indy = np.round((y - yp[0])/dy)
    indx = np.clip(fp_indx.astype(np.int), 0, Nxp - 1)
    indy = np.clip(fp_indy.astype(np.int), 0, Nyp - 1)
    f = fp[indy, indx]    
    valid_x = np.logical_and(fp_indx >= 0, fp_indx <= Nxp - 1)
    valid_y = np.logical_and(fp_indy >= 0, fp_indy <= Nyp - 1)
    valid = np.logical_and(valid_x, valid_y)
    f[np.logical_not(valid)] = 0
  else:
    raise ValueError(
      'Vrednost parametra "method" je lahko '
      '"linear" ali "nearest"!')

  f.shape = xshape

  return f

def interp3d(x, y, z, xp, yp, zp, fp, method='linear'):
  '''
  Interpolacija 3d slike. Izvedba funkcije temelji na uporabi for zank
  in je temu primerno počasna. Za hitro izvedbo interpolacije glej
  funkcijo map_coordinates modula scipy.ndimage ali funkcije modula interp.
  Funkcijske vrednosti izven definicijskega območja postavimo na 0.
  
  Parametri
  ----------
  x: numpy.ndarray
    x-koordinate točk, v katerih interpoliramo funkcijske vrednosti fp.
  y: numpy.ndarray
    y-koordinate točk, v katerih interpoliramo funkcijske vrednosti fp.
  z: numpy.ndarray
    z-koordinate točk, v katerih interpoliramo funkcijske vrednosti fp.
  xp: numpy.ndarray
    Vektor ekvidistantnih monotono naraščajočih x koordinat podatkovnih točk.
  yp: numpy.ndarray
    Vektor ekvidistantnih monotono naraščajočih y koordinat podatkovnih točk.
  zp: numpy.ndarray
    Vektor ekvidistantnih monotono naraščajočih z koordinat podatkovnih točk.
  fp: numpy.ndarray
    Funkcijske vrednosti v točkah, ki jih določajo xp, yp in zp. Velikost
    podatkovnega polja fp mora biti (zp.size, yp.size, xp.size).
  method: str
    Način interpolacije 'linear' (privzeto), ali 'nearest'.

  Vrne
  ----
  fxyz: numpy.ndarray
    Interpolirane funkcijske vrednosti v točkah f[z, y, x].

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> slika = imLoadRaw3d('./poglavje_1/mr_217x181x181_uint8.raw', 217, 181, 181)
  >>> xp = np.arange(0, 217, 1)
  >>> yp = zp = np.arange(0, 181, 1)
  >>> 
  >>> x = 100
  >>> y = z = np.arange(0, 181, 0.5)
  >>> Z, Y, X = np.meshgrid(z, y, x, indexing='ij')
  >>> prerez_nearest = np.squeeze(interp3d(X, Y, Z, xp, yp, zp, slika, 'nearest'))
  >>> prerez_linear = np.squeeze(interp3d(X, Y, Z, xp, yp, zp, slika))
  >>> 
  >>> pp.figure()
  >>> 
  >>> pp.subplot(1, 3, 1)
  >>> pp.title('Podatki')
  >>> pp.imshow(slika[:,:,100], cmap='gray')
  >>> 
  >>> pp.subplot(1, 3, 2)
  >>> pp.imshow(prerez_nearest, cmap='gray')
  >>> pp.title('Interpolacija z najbližjim sosedom')
  >>> 
  >>> pp.subplot(1, 3, 3)
  >>> pp.imshow(prerez_linear, cmap='gray')
  >>> pp.title('Trilinearn interpolacija')  
  '''
  x, y, z = np.asarray(x, np.float), np.asarray(y, np.float), \
    np.asarray(z, np.float)
  xp, yp, zp = np.asarray(xp, np.float), np.asarray(yp, np.float), \
    np.asarray(zp, np.float)
  fp = np.asarray(fp, np.float)
  
  xshape = x.shape
  x, y, z = x.flatten(), y.flatten(), z.flatten()
  dx = float(xp[1] - xp[0])
  dy = float(yp[1] - yp[0])
  dz = float(zp[1] - zp[0])
  Nxp, Nyp, Nzp = xp.size, yp.size, zp.size
  Nx, Ny, Nz = x.size, y.size, z.size
  if Nx != Ny or Nx != Nz:
    raise ValueError(
      'Število elementov v x, y in z mora biti enako!')


  f = np.zeros([Nx])
  if method == 'linear':
    # s for zanko
    '''
    for i in range(Nx):
      fp_indx = (x[i] - xp[0])/dx
      fp_indy = (y[i] - yp[0])/dy
      fp_indz = (z[i] - zp[0])/dz
      
      w_dx = fp_indx - np.floor(fp_indx)
      h_dy = fp_indy - np.floor(fp_indy)
      d_dz = fp_indz - np.floor(fp_indz)
      indx = np.clip(fp_indx.astype(np.int), 0, Nxp - 1)
      indx_plus_1 = np.clip(indx + 1, 0, Nxp - 1)
      indy = np.clip(fp_indy.astype(np.int), 0, Nyp - 1)
      indy_plus_1 = np.clip(indy + 1, 0, Nyp - 1)
      indz = np.clip(fp_indz.astype(np.int), 0, Nzp - 1)
      indz_plus_1 = np.clip(indz + 1, 0, Nzp - 1)
      
      a = (1.0 - w_dx)*(1.0 - h_dy)
      b = w_dx*(1.0 - h_dy)
      c = (1.0 - w_dx)*h_dy
      d = w_dx*h_dy

      b_dz = d_dz
      a_dz = 1.0 - d_dz
      
      fz_left = a*fp[indz, indy, indx] + \
                b*fp[indz, indy, indx_plus_1] + \
                c*fp[indz, indy_plus_1, indx] + \
                d*fp[indz, indy_plus_1, indx_plus_1]
      
      fz_right = a*fp[indz_plus_1, indy, indx] + \
                 b*fp[indz_plus_1, indy, indx_plus_1] + \
                 c*fp[indz_plus_1, indy_plus_1, indx] + \
                 d*fp[indz_plus_1, indy_plus_1, indx_plus_1]
        
      f[i] = fz_left*a_dz + fz_right*b_dz
      '''

      # brez for zanke
    fp_indx = (x - xp[0])/dx
    fp_indy = (y - yp[0])/dy
    fp_indz = (z - zp[0])/dz
      
    w_dx = fp_indx - np.floor(fp_indx)
    h_dy = fp_indy - np.floor(fp_indy)
    d_dz = fp_indz - np.floor(fp_indz)
    indx = np.clip(fp_indx.astype(np.int), 0, Nxp - 1)
    indx_plus_1 = np.clip(indx + 1, 0, Nxp - 1)
    indy = np.clip(fp_indy.astype(np.int), 0, Nyp - 1)
    indy_plus_1 = np.clip(indy + 1, 0, Nyp - 1)
    indz = np.clip(fp_indz.astype(np.int), 0, Nzp - 1)
    indz_plus_1 = np.clip(indz + 1, 0, Nzp - 1)
      
    a = (1.0 - w_dx)*(1.0 - h_dy)
    b = w_dx*(1.0 - h_dy)
    c = (1.0 - w_dx)*h_dy
    d = w_dx*h_dy

    b_dz = d_dz
    a_dz = 1.0 - d_dz

    fz_left = a*fp[indz, indy, indx] + \
              b*fp[indz, indy, indx_plus_1] + \
              c*fp[indz, indy_plus_1, indx] + \
              d*fp[indz, indy_plus_1, indx_plus_1]
      
    fz_right = a*fp[indz_plus_1, indy, indx] + \
               b*fp[indz_plus_1, indy, indx_plus_1] + \
               c*fp[indz_plus_1, indy_plus_1, indx] + \
               d*fp[indz_plus_1, indy_plus_1, indx_plus_1]

    f = fz_left*a_dz + fz_right*b_dz
    valid_x = np.logical_and(fp_indx >= 0, fp_indx <= Nxp - 1)
    valid_y = np.logical_and(fp_indy >= 0, fp_indy <= Nyp - 1)
    valid_z = np.logical_and(fp_indz >= 0, fp_indz <= Nzp - 1)
    valid_xy = np.logical_and(valid_x, valid_y)
    valid = np.logical_and(valid_xy, valid_z)
    f[np.logical_not(valid)] = 0

  elif method == 'nearest':
    # s for zanko
    '''
    for i in range(Nx):
      fp_indx = (x[i] - xp[0])/dx
      fp_indy = (y[i] - yp[0])/dy
      fp_indz = (z[i] - zp[0])/dz
      indx = np.clip(fp_indx.astype(np.int), 0, Nxp - 1)
      indy = np.clip(fp_indy.astype(np.int), 0, Nyp - 1)
      indz = np.clip(fp_indz.astype(np.int), 0, Nzp - 1)
      if fp_indx >= 0 and fp_indx <= Nxp - 1 and \
          fp_indy >= 0 and fp_indy <= Nyp - 1 and \
          fp_indz >= 0 and fp_indz <= Nzp - 1:
        f[i] = fp[indz, indy, indx] 
    '''

    # brez for zanke
    fp_indx = np.round((x - xp[0])/dx)
    fp_indy = np.round((y - yp[0])/dy)
    fp_indz = np.round((z - zp[0])/dz)
    indx = np.clip(fp_indx.astype(np.int), 0, Nxp - 1)
    indy = np.clip(fp_indy.astype(np.int), 0, Nyp - 1)
    indz = np.clip(fp_indz.astype(np.int), 0, Nzp - 1)
    f = fp[indz, indy, indx]    
    valid_x = np.logical_and(fp_indx >= 0, fp_indx <= Nxp - 1)
    valid_y = np.logical_and(fp_indy >= 0, fp_indy <= Nyp - 1)
    valid_z = np.logical_and(fp_indz >= 0, fp_indz <= Nzp - 1)
    valid_xy = np.logical_and(valid_x, valid_y)
    valid = np.logical_and(valid_xy, valid_z)
    f[np.logical_not(valid)] = 0
  else:
      raise ValueError(
        'Vrednost parametra "method" je lahko '
        '"linear" ali "nearest"!')

  f.shape = xshape

  return f

def imGrid2d(width, height, dx=1, dy=1, xoffset=0, yoffset=0):
  '''
  Ustvari koordinatno mrežo točk 2d slike, ki pripada podatkovnemu polju
  velikosti (depth, height, width).

  Parametri
  ---------
  width: int
    Širina slike izražena v slikovnih elementih (v smeri x koordinatne osi).
  height: int
    Višina slike izražena v slikovnih elementih (v smeri y koordinatne osi).
  dx: float
    Velikost slikovnega elemnta v smeri x koordinatne osi.
  dy: float
    Velikost slikovnega elemnta v smeri y koordinatne osi.
  xoffset: float
    X koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra 
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.
  yoffset: float
    Y koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra 
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.

  Vrne
  ----
  Y: numpy.ndarray
    Podatkovno polje Y koordinatnih točk slikovnih elementov 2d slike.
  X: numpy.ndarray
    Podatkovno polje X koordinatnih točk slikovnih elementov 2d slike.
    
  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> Y, X = imGrid2d(10, 5)
  >>> Yc, Xc = imGrid2d(10, 5, xoffset='center', yoffset='center')
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 3, 1)
  >>> pp.title('Brez zamika')
  >>> pp.plot(X, Y, 'xr')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> pp.title('Koordinatno izhodišče v težišču')
  >>> pp.plot(Xc, Yc, 'og')
  >>>
  >>> pp.subplot(1, 3, 3)
  >>> pp.title('Združeno')
  >>> pp.plot(X, Y, 'xr')
  >>> pp.plot(Xc, Yc, 'og')
  '''
  if yoffset == 'center':
    yoffset = -0.5*(height - 1)*dy
  if xoffset == 'center':
    xoffset = -0.5*(width - 1)*dx

  x = np.arange(xoffset, xoffset + width*dx ,dx, dtype=np.float)
  y = np.arange(yoffset, yoffset + height*dy ,dy, dtype=np.float)
  return np.meshgrid(y, x, indexing='ij')

def imGrid3d(width, height, depth, dx=1, dy=1, dz=1,
             xoffset=0, yoffset=0, zoffset=0):
  '''
  Ustvari koordinatno mrežo točk 3d slike, ki pripada podatkovnemu polju
  velikosti (depth, height, width).

  Parametri
  ---------
  width: int
    Širina slike izražena v slikovnih elementih (v smeri x koordinatne osi).
  height: int
    Višina slike izražena v slikovnih elementih (v smeri y koordinatne osi).
  depth: int
    Višina slike izražena v slikovnih elementih (v smeri z koordinatne osi).
  dx: float
    Velikost slikovnega elemnta v smeri x koordinatne osi.
  dy: float
    Velikost slikovnega elemnta v smeri y koordinatne osi.
  dz: float
    Velikost slikovnega elemnta v smeri z koordinatne osi.
  xoffset: float
    X koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra 
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.
  yoffset: float
    Y koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra 
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v
    razpolovišče pripadajoče stranice slike.
  zoffset: float
    Z koordinata slikovnega elementa na naslovu [0, 0, 0]. Vrednost parametra 
    je lahko tudi niz 'center', ki prestavi koordinatno izhodišče v 
    razpolovišče pripadajoče stranice slike.

  Vrne
  ----
  Z: numpy.ndarray
    Podatkovno polje Z koordinatnih točk slikovnih elementov 3d slike.
  Y: numpy.ndarray
    Podatkovno polje Y koordinatnih točk slikovnih elementov 3d slike.
  X: numpy.ndarray
    Podatkovno polje X koordinatnih točk slikovnih elementov 3d slike.

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>> from mpl_toolkits.mplot3d import Axes3D
  >>>
  >>> Z, Y, X = imGrid3d(20, 10, 5)
  >>> Zc, Yc, Xc = imGrid3d(20, 10, 5, xoffset='center', yoffset='center', zoffset='center')
  >>>
  >>> fig = pp.figure()
  >>>
  >>> ax = fig.add_subplot(131, projection='3d')
  >>> ax.plot(X.flatten(), Y.flatten(), Z.flatten(), 'xr')
  >>> ax.set_title('Brez zamika')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> ax = fig.add_subplot(132, projection='3d')
  >>> ax.plot(Xc.flatten(), Yc.flatten(), Zc.flatten(), 'og')
  >>> ax.set_title('Koordinatno izhodišče v težišču')
  >>>
  >>> pp.subplot(1, 3, 3)
  >>> ax = fig.add_subplot(133, projection='3d')
  >>> ax.plot(X.flatten(), Y.flatten(), Z.flatten(), 'xr')
  >>> ax.plot(Xc.flatten(), Yc.flatten(), Zc.flatten(), 'og')
  >>> ax.set_title('Združeno')
  '''

  if zoffset == 'center':
    zoffset = -0.5*(depth - 1)*dz
  if yoffset == 'center':
    yoffset = -0.5*(height - 1)*dy
  if xoffset == 'center':
    xoffset = -0.5*(width - 1)*dx

  x = np.arange(xoffset, xoffset + width*dx ,dx, dtype=np.float)
  y = np.arange(yoffset, yoffset + height*dy ,dy, dtype=np.float)
  z = np.arange(zoffset, zoffset + depth*dz ,dz, dtype=np.float)
  return np.meshgrid(z, y, x, indexing='ij')

def _conv2d(data, kernel):
  '''
  Konvolucija 2d slike s poljubnim jedrom na podlagi funkcije convolve
  modula scipy.ndimage.
  
  Parametri
  ---------  
  data: numpy.ndarray
    Podatkovno polje slike.
  kernel: numpy.ndarray
    Konvolucijsko jedro.
  Vrne
  ----
  conv: numpy.ndarray
    Konvolucija med vhodno sliko in konvolucijskim jedrom.
    
  Opombe
  ------
  Funkcijske vrednosti izven podatkovnega polja postavimo na 0.
  Funkcija scipy.ndimage.convolve zahteva veliko pomnilnika pri velikih
  konvolucijskih jedrih. Učinkovitejša alternativa je funkcija
  scipy.signal.convolve2d!
    
  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> konv_jedro = np.ones([5, 5])
  >>> konv_jedro /= konv_jedro.sum()
  >>>
  >>> slika = imLoadRaw2d('./poglavje_1/mr_217x181_uint8.raw', 217, 181)
  >>> filtrirana = conv2d(slika, konv_jedro)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 2, 1)
  >>> pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Slika')
  >>>
  >>> pp.subplot(1, 2, 2)
  >>> pp.imshow(filtrirana, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Filtrirana slika')
  '''

  # konvolucija z uporabo štirih for zank  
  '''
  K = np.asarray(kernel, np.float)    
  S = np.asarray(data, np.float)
  c = np.floor(np.array(K.shape)/2.0) # naslov središča konvolucijskega jedra
  H, W = S.shape	# velikost vhodne slike
  A, B = K.shape	# velikost konvolucijskega jedra
  Sc = np.zeros([H, W])	# podatkovno polje izhodne (zglajene) slike
  for i in range(H):		# vrstica slikovnega elementa
    for j in range(W):	# stolpec slikovnega elementa
      for k in range(A):  
        for l in range(B):
          # Premik naslovov.
          ic = i - (k - c[0])	
          jc = j - (l - c[1])	
          # Pred izračunom preverimo veljavnost naslova slikovnega elementa.
          if ic > 0 and ic < H and jc > 0 and jc < W:
            Sc[i, j] += K[k, l]*S[ic, jc]
  return Sc
  '''
  # hitra konvolucija s funkcijo convolve modula scipy.ndimage
  # ndimage convolve neučinkovit pri velikih konvolucijskih jedrih 
  # (porabi ogromno pomnilnika)! Alternativa je scipy.signal.convolve2d
  # return ndim.convolve(data, kernel, **kwargs)
  return convolve2d(data, kernel, mode='same')

def imDecimate2d(img, kernel=None, level=1):
  '''
  Zaporedoma decimira sliko s faktorjem 2 (piramidna shema).

  Parametri
  ---------
  img: numpy.ndarray
    Slika za decimacijo.
  kernel: numpy.ndarray
    Nizkoprepustni jedro decimacijskega filtra.
  level: int
    Celoštevilska stopnja decimacije. Velikost slike v vsaki razsežnosti se
    zmanjša za faktor 2**level.

  Vrne
  ----
  imdec: numpy.ndarray
    Decimirana slika.

  Opombe
  ------
  S funkcijske je mogoče udejaniti zgolj decimacijo slik s potencami števila 2.

  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> slika = imLoadRaw2d('./poglavje_1/mr_217x181_uint8.raw', 217, 181)
  >>> filtrirana = imDecimate2d(slika)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 5, 1)
  >>> pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Slika')
  >>>
  >>> for i in range(4):
  ...   slika = imDecimate2d(slika)
  ...   pp.subplot(1, 5, i + 2)
  ...   pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
  ...   pp.title('Decimirana slika k={}'.format(2**(i + 1)))
  >>>
  '''
  if kernel is None:
    kernel = np.array([[1.0/16, 1/8, 1/16],
                       [1/8,    1/4, 1/8],
                       [1.0/16, 1/8, 1/16]])
  oimg = np.array(img, 'float')
  for i in range(level):
    oimg = conv2d(oimg, kernel)
    oimg = oimg[::2, ::2]

  return oimg

def imDecimateGs(img, kernel=None, level=1):
    return imDecimate2d(img, kernel, level).astype(img.dtype)

def imDecimateRGB(img, kernel=None, level=1):
    r = imDecimate2d(img[:,:,0], kernel, level)
    g = imDecimate2d(img[:,:,1], kernel, level)
    b = imDecimate2d(img[:,:,2], kernel, level)

    r = r.astype(img.dtype)
    g = g.astype(img.dtype)
    b = b.astype(img.dtype)

    return np.dstack([r,g,b])
#%% Poglavje 3

def srgb2gs(img):
  '''
  Pretvori barvno SRGB sliko v sivinsko. Pretvorba se izvede kot utežena
  vsota posameznih barvnih komponent SRG slike, in sicer kot
  
  0.299*(rdeča komponenta) + 0.587*(zelena komponenta) + 0.144*(modra komponenta).
    
  Parametri
  ---------
  img: numpy.ndarray
    Barvna slika kot numpy podatkovno polje velikosti (višina, širina, 3).
    
  Vrne
  ----
  gsimg: numpy.ndarray
    Sivinska slika 0.299*img[:,:,0] + 0.587*img[:,:,1] + 0.144*img[:,:,2].
      
  Primeri uporabe
  --------------
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>> import numpy as np
  >>>
  >>> rgb = np.array(im.open('./poglavje_3/ColorChecker.jpg'))
  >>> gs = srgb2gs(rgb)
  >>>
  >>> pp.figure()
  >>>  
  >>> pp.subplot(1, 2, 1)
  >>> pp.imshow(rgb)
  >>> pp.title('Barvna slika')
  >>>  
  >>> pp.subplot(1, 2, 2)
  >>> pp.imshow(gs, cmap='gray')
  >>> pp.title('Sivinska slika')
  >>>
  '''
  return 0.299*img[:,:,0] + 0.587*img[:,:,1] + 0.144*img[:,:,2]


def imHistogram(img, bins=256, span=None, density=False, title=None):
  '''
  Izračuna histogram vhodne slike. Ko je vrednost parametra title različna
  od None, hiostogram prikaže v trenutnem grafičnem (pod)oknu.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  bins: int
    Število razredov histograma.
  span: list
    Razpon vrednosti kot (spodnja, zgornja). Če je vrednost None, se
    spodnja in zgornja meja določita samodejno kot najmanjša in največja
    sivinska vrednost v sliki.
  density: bool
    Če je vrednost True, se histogram normalizira, tako da predstavlja oceno
    gostote verjetnosti (vsota vseh normaliziranih frekvenc je enaka 1).
  title: str
    Naslovna vrstica grafičnega prikaza. Če je vrednost None, se histogram ne
    izriše.

  Vrne
  ----
  f: numpy.ndarray
    Frekvence razredov.
  edges: numpy.ndarray
    Robne točke razredov

  Primeri uporabe
  --------------
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>> import numpy as np
  >>>
  >>> img = imLoadRaw2d('./poglavje_1/rezultati/mr_256x256_uint8.raw', 256, 256)
  >>> 
  >>> pp.figure()
  >>> pp.subplot(1, 2, 1)
  >>> pp.title("Slika")
  >>> pp.imshow(img, cmap='gray')
  >>>
  >>> pp.subplot(1, 2, 2)
  >>> f, e = imHistogram(img, 64, title="Histogram slike")
  >>>
  '''
  img = np.asarray(img)
  if span is None:
    span = (img.min(), img.max())
  span = np.asarray(span, np.float)
  edges = np.linspace(span[0], span[1], bins)
  hist = np.zeros([bins])
  for i in range(bins - 1):
    hist[i] = np.count_nonzero(np.logical_and(
      img >= edges[i], img < edges[i + 1]))
  hist[-1] = np.count_nonzero(np.logical_and(
    img >= edges[-2], img <= edges[-1]))

  if density:
    hist /= hist.sum()

  if title is not None:
    pp.bar(edges, hist, 0.8*(span[1] - span[0])/bins)
    pp.title(title)
        
  return hist, edges

def ssim(imgx, imgy, l=255, n=11, sigma=1.5,
         k1=0.01, k2=0.03):
  '''
  Izračuna vrednost metrike strukturne podobnosti (ang. Structural Similarity 
  Index - SSIM) za vhodni sliki.

  Parametri
  ---------
  imgx: numpy.ndarray
    Prva vhodna slika.
  imgy: numpy.ndarray
    Druga vhodna slika
  l: int
    Dinamično območje sivinskih vrednosti vhodnih slik.
  sigma: float
    Parameter sigma Gaussove funkcije s katero utežimo sivinske vrednosti
    v izbrani okolici slikovnega elementa.
  n: int
    Velikost okolice slikovnega elementa, ki jo uporabi za izračun SSIM.
  k1, k2: float
    Parametra metrike SSIM.

  Vrne
  ----
  ossim: numpy.ndarray
    Vrednosti SSIM (podatkovno polje velikosti vhodne slike). 

  Opombe
  ------
  Velikosti vhodnih slik imga in imgb morata biti enaki.
  '''
  imgx = np.asarray(imgx, dtype=np.float)
  imgy = np.asarray(imgy, dtype=np.float)

  n_half = int(n//2)
  n = n_half*2 + 1
  if imgx.shape != imgy.shape:
    raise ValueError('Velikosti vhodnih slik "imgx" in "imgy" '
                     'morata biti enaki!')

  x = y = np.arange(-n_half, n_half + 1, dtype=np.float)
  Y, X = np.meshgrid(y, x, indexing='ij')
  W = 1.0/(2.0*np.pi*sigma**2)* \
      np.exp(-(X**2 + Y**2)/(2.0*sigma**2))
  W *= 1.0/W.sum()

  imgxp = imPad2d(imgx, n_half, boundary='reflect')
  imgyp = imPad2d(imgy, n_half, boundary='reflect')

  c1 = (k1*l)**2
  c2 = (k2*l)**2

  H, W = imgx.shape
  ossim = np.zeros(imgx.shape)
  for i in range(W):
    for j in range(H):
      xr = W*imgxp[j: j + n, i: i + n]
      yr = W*imgyp[j: j + n, i: i + n]

      mxr = xr.mean()
      myr = yr.mean()
      sxr = xr.std()
      syr = yr.std()
      sxyr = (xr*yr).mean() - mxr*myr

      ossim[j, i] = (2.0*mxr*myr + c1)*(2.0*sxyr + c2)/ \
                    ((mxr**2 + myr**2 + c1)*(sxr**2 + syr**2 +c2))

  return ossim
#%% Poglavje 4

def imPad2d(img, n, boundary="constant", fillvalue=0):
  '''
  Razširi definicijsko območje slike s pasom dane širine.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  n: list, int
    Pas razširitve v x (n[1]) in y smeri (n[0]). Če n ni seznam, potem
    se v smeri obeh koordinatnih osi uporabi ista vrednost.
  boundary: str
    Eden izmed načinov razširitve podatkovnega polja, in sicer:
    
        'constant' - razširitev z vrednostjo parametra value
        
        'edge'     - razširitev z najbližjim slikovnim elementom vhodne slike

        'reflect'  - razširitev z zrcaljenjem vhodne slike

  fillvalue: int, float
    Vrednost reazširjenega dela podatkovnega polja pri načinu 'const'.

  Vrne
  ----
  pimg: numpy.ndarray
    Razširjena vhodna slika.
  
  Primeri uporabe:
  ----------------
  >>> import numpy as np
  >>>
  >>> slika = np.array([[1,2,3],[4,5,6],[7,8,9]])
  >>> imPad2d(slika, 2, 'constant', fillvalue=100)
  >>> imPad2d(slika, 2, 'edge')
  >>> imPad2d(slika, 2, 'reflect')
  >>>
  '''
  boundary = str(boundary).lower()
  # primer kode za razširite s konstantno ali najbližjo vrednostjo
  '''
  if isinstance(n, int)
    py, px = n, n
  else:
    py, px = n[0], n[1]
    
  H, W = img.shape[0], img.shape[1]
  Hp, Wp = H + 2*py, W + 2*px
  if boundary == 'nearest':
    oimg = np.zeros([Hp, Wp], dtype=img.dtype)    
    oimg[-py:, px:-px] = img[-1, :]
    oimg[py:-py, :px] = img[:,0].reshape(H, 1)
    oimg[py:-py:, -px:] = img[:,-1].reshape(H, 1)

    oimg[:py, :px]=img[0,0]
    oimg[:py, -px:]=img[0, -1]
    oimg[-py:, :px]=img[-1,0]
    oimg[-py:, -px:]=img[-1,-1]
  elif boundary == 'constant':
    oimg = np.tile(img.dtype.type(value), [Hp, Wp])
		
  oimg[py:-py, px:-px] = img
  return oimg
  '''
  # učinkovitejša rešitev s funkcijo pad modula numpy
  if boundary == 'constant':
      return np.pad(img, n, mode=boundary, constant_values=fillvalue)

  elif boundary in ['reflect', 'wrap']:
      return np.pad(img, n, mode=boundary)

  elif boundary == 'nearest':
      return np.pad(img, n, mode='edge')

  elif boundary == 'mirror':
      return np.pad(img, n, mode='symmetric')

  else:
    raise ValueError(
      'Vrednost parametra "mode" je lahko '
      '"constant", "reflect", "nearest" ali "mirror"!')

def conv1d(data, kernel, boundary='nearest', fillvalue=0):
  '''
  1d konvolucija signala s konvolucijskim jedrom.
  
  Parametri
  ---------
  data: numpy.ndarray
    Vhodno podatkovno polje.
  kernel: numpy.ndarray
      Konvolucijsko jedro.
  boundary: str
    Način obravnave funkcijskih vrednosti izven definicijskega območja, in 
    sicer eden izmed 'reflect', 'constant', 'nearest', 'mirror' ali 'wrap'.
  fillvalue: int, float
    Vrednosti izven definicijskega območja podatkovnega polja
    pri razširitvi s konstantno vrednostjo mode = 'constant'.
      
  Vrne
  ----
  fdata: numpy.ndarray
    Rezultat konvolucije vhodnih podatkov z danim jedrom. Velikost vrnjenega
    podatkovnega polja je enaka velikosti vhodnega podatkovnega polja data.
     
  Primeri uporabe:
  ----------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> f = np.tile([0.0, 0, 0, 0, 0, 0, 1, 1, 1 ,1, 1, 1], [5])
  >>> k = [1.0/3, 1.0/3, 1.0/3]
  >>> fc = conv1d(f, k)
  >>>
  >>> pp.figure()
  >>>  
  >>> pp.subplot(1, 2, 1)
  >>> pp.plot(f)
  >>> pp.title('Izvirni signal')
  >>>  
  >>> pp.subplot(1, 2, 2)
  >>> pp.plot(fc)
  >>> pp.title('Signal po konvoluciji')
  '''
  # celotna izvedba s Python kodo
  '''
  mode = str(mode).lower()
  a = len(kernel)
  c = p = np.floor(a/2)
  N = data.size
  Np = N + 2*p
  pdata = np.zeros([Np])
  pdata[p:-p] = data
  if mode == 'nearest':
    pdata[:p] = data[0]
    pdata[-p:] = data[-1]
  elif mode == 'constant':
    pdata[:p] = value
    pdata[-p:] = value
  else:
    raise ValueError('Vrednost parametra "mode" je lahko '
                     '"nearest" ali "constant"!')
  odata = np.zeros([N])
  for i in range(N):
    for j in range(a):
      odata[i] += kernel[j]*pdata[p + i - (j - c)] 
  return odata.reshape(data.shape)
  '''
  # učinkovitejša rešitev s funkcijo convolve1d modula scipy.ndimage
  return ndim.convolve1d(data, kernel, mode=boundary, cval=fillvalue)

def conv2d(data, kernel, boundary='constant', fillvalue=0):
  '''
  Hitra konvolucija 2d slike s poljubnim jedrom na podlagi funkcije convolve
  modula scipy.ndimage.
  
  Parametri
  ---------  
  data: numpy.ndarray
    Podatkovno polje slike.
  kernel: numpy.ndarray
    Konvolucijsko jedro.
  boundary: str
    Način obravnave sivinskih vrednosti izven definicijskega območja slike.
  fillvalue: int, float
    Sivinska vrednost izven definicijskega območja slike pri vrednosti
    parametra boundarz="constant".
  
  Vrne
  ----
  conv: numpy.ndarray
    Konvolucija med vhodno sliko in konvolucijskim jedrom.
    
  Opombe
  ------
  Funkcijske vrednosti izven podatkovnega polja postavimo na 0.
  Funkcija scipy.ndimage.convolve zahteva veliko pomnilnika pri velikih
  konvolucijskih jedrih. Učinkovitejša alternativa je funkcija
  scipy.signal.convolve2d!
    
  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> konv_jedro = np.ones([5, 5])
  >>> konv_jedro /= konv_jedro.sum()
  >>>
  >>> slika = imLoadRaw2d('./poglavje_1/mr_217x181_uint8.raw', 217, 181)
  >>> filtrirana = conv2d(slika, konv_jedro)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 2, 1)
  >>> pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Slika')
  >>>
  >>> pp.subplot(1, 2, 2)
  >>> pp.imshow(filtrirana, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Filtrirana slika')
  '''
  n = np.floor(np.array(kernel.shape)/2.0).astype(np.int)

  pdata = imPad2d(data, n, boundary, fillvalue)
  odata = _conv2d(pdata, kernel)

  return odata[n[0]:-n[0], n[1]:-n[1]]

def gaussianKernel1d(sigma, truncate=4):
  '''
  Ustvari 1D Gaussovo jedro standardne deviacije sigma in širine +- 3*sigma.
  
  Parametri
  ---------
  sigma: float
    Standardna deviacija Gaussovega jedra.
  truncate: float
    Predpostavimo, da je vrednost Gaussove funkcije izven intervala 
    [-truncate*sigma, truncate*sigma] enaka 0.
    
  Vrne
  ----
  kernel: numpy.ndarray
    Gaussovo konvolucijsko jedro jedro.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> k = gaussianKernel1d(3)
  >>>
  >>> pp.figure()
  >>> pp.plot(k)
  >>> pp.title('Gaussovo jedro, sigma=3')
  >>>
  '''
  x = np.arange(2*np.ceil(truncate*sigma) + 1)
  x -= x.mean()
  kernel = 1.0/(sigma*np.sqrt(2.0*np.pi))* \
    np.exp(-x**2*(1.0/(2.0*sigma**2)))
  kernel /= kernel.sum()
  return kernel

def gaussianKernel2d(sigma, truncate=4):
  '''
  Ustvari 2D Gaussovo jedro standardne deviacije sigma.
  
  Parametri
  ---------
  sigma: float
    Standardna deviacija Gaussovega jedra.
  truncate: float
    Predpostavimo, da je vrednost Gaussove funkcije izven intervala 
    [-truncate*sigma, truncate*sigma] enaka 0.
  
  Vrne
  ----
  kernel: numpy.ndarray
    2D Gaussovo konvolucijsko jedro.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from mpl_toolkits.mplot3d import Axes3D
  >>>
  >>> k = gaussianKernel2d(3)
  >>> Y, X = imGrid2d(k.shape[1], k.shape[0]) 
  >>> fig = pp.figure()
  >>> ax = fig.add_subplot(111, projection='3d')
  >>> ax.plot_wireframe(X, Y, k, color='r')
  >>> pp.title('2D Gaussovo jedro, sigma=3')
  >>>
  '''  
  n = int(2*np.ceil(truncate*sigma) + 1)
  Y, X = imGrid2d(n, n, xoffset='center', yoffset='center')
  kernel = 1.0/(2.0*np.pi*sigma**2)* \
    np.exp(-(X**2 + Y**2)*(1.0/(2.0*sigma**2)))
  kernel /= kernel.sum()
  return kernel

def imGaussFilt2d(img, sigma, truncate=4, boundary='constant', fillvalue=0):
  '''
  2D filtriranje z Gaussovim jedrom, ki se izvede zaporedoma, najprej
  po vrsticah, nato pa po stolpcih slike.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  sigma: float
    Standardna deviacija Gaussovega jedra.
  truncate: float
    Predpostavimo, da je vrednost Gaussove funkcije izven intervala 
    [-truncate*sigma, truncate*sigma] enaka 0.    
  mode: str
    Način obravnave funkcijskih vrednosti izven definicijskega območja, in 
    sicer eden izmed 'reflect', 'constant', 'nearest', 'mirror' ali 'wrap'.
  cval: int, float
    Vrednosti izven definicijskega območja podatkovnega polja
    pri razširitvi s konstantno vrednostjo mode = 'constant'.
    
  Vrne
  ----
  fimg: numpy.ndarray
    Vhodno sliko zglajeno z Gaussovim filtrom.
    
  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> img = imLoadRaw2d('./vaja3/vaja3_350x350_uint8.raw', 350, 350)
  >>> fimg = imGaussFilt2d(img, 5)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 2, 1)
  >>> pp.imshow(img, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Izvirna slika')
  >>>
  >>> pp.subplot(1, 2, 2)
  >>> pp.imshow(fimg, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Filtrirana slika, sigma=5')
  >>>
  '''
  # izvedba z zaporednima 1D konvolucijama
  '''
  mode = str(mode).lower()
  H, W = img.shape[0], img.shape[1]
  K = gaussianKernel1d(sigma)
  oimg = np.zeros(img.shape)
  # Filtriramo po vrsticah slike.
  for i in range(H):
    oimg[i, :] = conv1d(img[i, :], K, boundary, fillvalue)
    # Filtriramo po stolpcih slike.
  for i in range(W):
    oimg[:, i] = conv1d(oimg[:, i], K, boundary, fillvalue)
  return oimg
  '''
  # učinkovitejša izvedba konvolucije s funkcijo gaussian_filter
  return ndim.gaussian_filter(img, sigma, mode=boundary, cval=fillvalue)


def imSharpen2d(img, kind='mask', c=1.0, sigma=1.0, **kwargs):
  '''
  Ostrenje slike z Laplaceovim operatorjem ali maskiranje.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika
  kind: str
    Način ostrenja. Uporabi 'laplace' za Laplaceovo ali 'mask' za ostrenje
    z maskiranjem.
  c: float
    Parameter ostrenja z laplaceovim operatorjem.
  sigma: float
    Standardna deviacija Gaussovega filtra pri ostrenju z maskiranjem.
  mode: str
    Način obravnave funkcijskih vrednosti izven definicijskega območja, in 
    sicer eden izmed 'reflect', 'constant', 'nearest', 'mirror' ali 'wrap'.
  cval: int, float
    Vrednosti izven definicijskega območja podatkovnega polja
    pri razširitvi s konstantno vrednostjo mode = 'constant'.
    
  Vrne
  ----
  simg: numpy.ndarray
    Izostrena slika.
    
  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> img = imLoadRaw2d('./vaja3/vaja3_350x350_uint8.raw', 350, 350)
  >>> mimg = np.clip(imSharpen2d(img, 'mask'), 0, 255)
  >>> limg = np.clip(imSharpen2d(img, 'laplace'), 0, 255)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 3, 1)
  >>> pp.imshow(img, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Izvirna slika')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> pp.imshow(mimg, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Ostrenje z maskiranjem')
  >>> 
  >>> pp.subplot(1, 3, 3)
  >>> pp.imshow(limg, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Ostrenje z Laplaceom')
  >>>
  '''
  img = img.astype(np.float)
  if kind == 'mask':
    M = img - imGaussFilt2d(img, sigma, **kwargs)
    oimg = img + c*M
  elif kind == 'laplace':
    Lk = np.array([[0, 1, 0],[1, -4, 1],[0, 1, 0]], np.float)
    oimg = img - c*conv2d(img, Lk, **kwargs)
  else:
    raise ValueError('Vrednost parametra "kind" je lahko '
                     '"mask" ali "laplace"!')
  return oimg

def imStatFilt2d(img, n=3, kind='median',
                 boundary='reflect', fillvalue=0.0):
  '''
  Statistično filtriranje z minimalno, maksimalno vrednostjo ter mediano.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika
  n: int
    Velikost filtrskega jedra.
  kind: str
    Način filtriranja. Uporabi 'min' za minimalno, 'max' za maksimalno,
    in 'median' za filtriranje z mediano.
  boundary: str
    Način obravnave funkcijskih vrednosti izven definicijskega območja, in 
    sicer eden izmed 'reflect', 'constant', 'nearest', 'mirror' ali 'wrap'.
  fillvalue: int, float
    Vrednosti izven definicijskega območja podatkovnega polja
    pri razširitvi s konstantno vrednostjo mode = 'constant'.
    
  Vrne
  ----
  fimg: numpy.ndarray
    Statistično filtrirana slika.

  Opombe
  ------
   Za učinkovitejšo izvedbo filtriranja uporabi funkcije minimum_filter
   maximum_filte in median_filte modula scipy.ndimage.
    
  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> img = imLoadRaw2d('./poglavje_4/ct_sp_350x350_uint8.raw', 350, 350)
  >>> medianimg = imStatFilt2d(img)
  >>> gaussimg = imGaussFilt2d(img, sigma=1)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 3, 1)
  >>> pp.imshow(img, cmap='gray')
  >>> pp.title('Izvirna slika')
  >>>
  >>> pp.subplot(1, 3, 2)
  >>> pp.imshow(medianimg, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Filriranje z mediano [3, 3]')
  >>> 
  >>> pp.subplot(1, 3, 3)
  >>> pp.imshow(gaussimg, cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Filtriranje z Gaussovim jedrom, sigma=1')
  >>>
  '''
  H, W = img.shape
  if kind == 'min': 
    f = np.min
  elif kind == 'max': 
    f = np.min
  elif kind == 'median': 
    f = np.median
  else:
    raise ValueError('Vrednost parametra kind je lahko '
                     '"min" "max" ali "median"!')
      
  p = int(np.floor(n/2.0))
  pimg = imPad2d(img, [p, p], boundary, fillvalue)
  oimg = np.zeros(img.shape)
  for i in range(H):
    for j in range(W):
      oimg[i, j] = f(pimg[p + i:p + i + n + 1, 
                          p + j:p + j + n + 1])
  return oimg

def imGaussFilt3d(img, sigma, boundary="constant", fillvalue=0):
  '''
  3D filtriranje z Gaussovim jedrom izvedeno z zaporednim filtriranjem
  z 1D Gaussovim jedrom.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  sigma: float
    Standardna deviacija Gaussovega jedra.
  boundary: str
    Način obravnave funkcijskih vrednosti izven definicijskega območja, in 
    sicer eden izmed 'reflect', 'constant', 'nearest', 'mirror' ali 'wrap'.
  fillvalue: int, float
    Vrednosti izven definicijskega območja podatkovnega polja
    pri razširitvi s konstantno vrednostjo mode = 'constant'.
    
  Vrne
  ----
  fimg: numpy.ndarray
    Filtrirana slika.
    
  Opombe
  ------
  Za učinkovitejšo izvedbo filtriranja uporabi funkcijo gaussian_filter
  modula scipy.ndimage.
  
  Primeri uporabe
  ---------------
  >>> from matplotlib import pyplot as pp
  >>>
  >>> img = imLoadRaw3d('./poglavje_2/mr_217x181x181_uint8.raw', 217, 181, 181)
  >>> fimg = imGaussFilt3d(img, 1)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(1, 2, 1)
  >>> pp.imshow(img[90], cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Prerez z=90 izvirne slike')
  >>>
  >>> pp.subplot(1, 2, 2)
  >>> pp.imshow(fimg[90], cmap='gray', vmin=0, vmax=255)
  >>> pp.title('Prerez z=90 filtrirane slike')
  >>>
  '''  
  H, W, D = img.shape
  K = gaussianKernel1d(sigma)
  oimg = np.zeros(img.shape)
  for i in range(D):
    for j in range(H):
      oimg[j, : ,i] = conv1d(img[j, :, i], K, boundary, fillvalue)
  for i in range(D):
    for j in range(W):
      oimg[:, j, i] = conv1d(oimg[:, j, i], K, boundary, fillvalue)
  for i in range(H):
    for j in range(W):
      oimg[i, j, :] = conv1d(oimg[i, j, :], K, boundary, fillvalue)
  return oimg

def surf(x, y, z, *args, fig=None, subplot=111, **kwargs):
  '''
  Izriše funkcijo dveh parametrtov z(x, y) kot ploskev.

  Parametri
  ---------
  x: numpy.ndarray ali None
    x koordinate točk, v katerih je definirana funkcijska vrednost z. Ko je
    vrednost parametra None, se koordinatna mreža točk ustvari s funkcijo
    meshgrid modula numpy, in sicer z vektorjem np.arange(z.shape[1]).
  y: numpy.ndarray ali None
    y koordinate točk, v katerih je definirana funkcijska vrednost z. Ko je
    vrednost parametra None, se koordinatna mreža točk ustvari s funkcijo
    meshgrid modula numpy, in sicer z vektorjem np.arange(z.shape[0]).
  z: numpy.ndarray
    Funkcijske vrednosti za vse pare točk, ki jih definirata 2D podatkovni
    polji x in y.
  fig: None ali matplotlib.figure.Figure
    Grafično okno.
  subplot: int, tuple
    Oznaka podokna.
    
  *args, **kwargs: poljubno
    Vrednosti parametrov se posredujejo funkciji plot_surface.

  Primeri uporabe
  ---------------
  >>> import bstlib
  >>> from matplotlib import pyplot as pp
  >>>
  >>> surf(None, None, bstlib.gaussianKernel2d(20), shade=True)
  >>>
  '''
  z = np.asarray(z)
  if x is None or y is None:
    y, x = np.meshgrid(np.arange(z.shape[0]), np.arange(z.shape[1]),
                      indexing='ij')

  if 'cstride' not in kwargs:
    kwargs['cstride'] = max(z.shape[0], z.shape[1])//30
  if 'rstride' not in kwargs:
    kwargs['rstride'] = max(z.shape[0], z.shape[1])//30

  if fig is None:
    fig = pp.gcf()

  if ifinstance(subplot, (tuple, list)):
    ax = fig.add_subplot (*subplot , projection='3d')
  else:
    ax = fig.add_subplot (subplot , projection='3d')
  ax.plot_surface (x, y, z, *args, **kwargs)

  return ax

def wireframe(x, y, z, *args, fig=None, subplot=111, **kwargs):
  '''
  Izriše funkcijo dveh parametrtov z(x, y) kot žični model.

  Parametri
  ---------
  x: numpy.ndarray ali None
    x koordinate točk, v katerih je definirana funkcijska vrednost z. Ko je
    vrednost parametra None, se koordinatna mreža točk ustvari s funkcijo
    meshgrid modula numpy, in sicer z vektorjem np.arange(z.shape[1]).
  y: numpy.ndarray ali None
    y koordinate točk, v katerih je definirana funkcijska vrednost z. Ko je
    vrednost parametra None, se koordinatna mreža točk ustvari s funkcijo
    meshgrid modula numpy, in sicer z vektorjem np.arange(z.shape[0]).
  z: numpy.ndarray
    Funkcijske vrednosti za vse pare točk, ki jih definirata 2D podatkovni
    polji x in y.
  fig: None ali matplotlib.figure.Figure
    Grafično okno.
  subplot: int ali tuple
    Oznaka podokna.
  *args: tuple
    Vrednosti parametrov se posredujejo funkciji plot_wireframe.
  **kwargs: dict
    Vrednosti parametrov se posredujejo funkciji plot_wireframe.

  Primeri uporabe
  ---------------
  >>> import bstlib
  >>> from matplotlib import pyplot as pp
  >>>
  >>> fig = pp.figure()
  >>> surf(None, None, z=bstlib.gaussianKernel2d(20), fig=fig, subplot=121, shade=True)
  >>> wireframe(None, None, z=bstlib.gaussianKernel2d(20), fig=fig, subplot=122)
  >>>
  '''
  z = np.asarray(z)
  if x is None or y is None:
    y, x = np.meshgrid(np.arange(z.shape[0]), np.arange(z.shape[1]),
                      indexing='ij')

  if 'cstride' not in kwargs:
    kwargs['cstride'] = max(z.shape[0], z.shape[1])//30
  if 'rstride' not in kwargs:
    kwargs['rstride'] = max(z.shape[0], z.shape[1])//30

  if fig is None:
    fig = pp.gcf()

  if ifinstance(subplot, (tuple, list)):
    ax = fig.add_subplot(*subplot , projection='3d')
  else:
    ax = fig.add_subplot (subplot , projection='3d')
  ax.plot_wireframe (x, y, z, *args, **kwargs)

  return ax

#%% Poglavje 5

def imShowEx(img, extent=None,
             title='', axis='off', subplot=None, **kwargs):
  '''
  Prikaže barvno ali sivinsko sliko, ki je definirana na poljubni mreži točk.

  Parametri
  ---------
  img: numpy.ndarray
    Barvna ali sivinska slika.
  extent: list
    Koordinate levega zgornjega in desnega spodnjega krajišča slike kot
    [xmin, xmax, ymin, ymax].
  title: str
    Naslovna vrstica grafičnega prikaza.
  axis: bool
    Vklopi 'on' ali izklopi 'off' prikaz koordinatnih osi.
  subplot: int, tuple
    Naslov podokna ali None.
  kwargs: dict
    Dodatni argumenti se prenesejo funkciji imshow.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_128x256_uint8.raw', 128, 256)
  >>> imShowEx(im, title='Slika - ne upošteva vel. slik. elementa', subplot=121)
  >>> imShowEx(im, [0, 255, 0, 255], 'Slika - vel. slik. el. 1x2 mm', subplot=122)
  >>>
  '''
  if subplot is not None:
    if isinstance(subplot, (list, tuple)):
      pp.subplot(*subplot)
    else:
      pp.subplot(subplot)

  if len(img.shape) <= 2:
    cmap='gray'

  pp.imshow(img.squeeze(), extent=extent, cmap=cmap, **kwargs)
  pp.title(title)
  pp.axis(axis)

def imScale(img, slope, intersection):
  '''
  Linearna preslikava sivinskih vrednosti po enačbi img*slope + intersection.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna referenčna slika.
  slope: float
    Strmina premice, ki določa preslikavo.
  intersection: float
    Presečišče premice, ki določa preslikavo, z y osjo.

  Vrne
  ----
  oimg: numpy.ndarray
    Linearno preslikane sivinske vrednosti slike.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_256x256_uint8.raw', 256, 256)
  >>> ims = imScale(im, 0.5, 128.0)
  >>>  
  >>> imShowEx(im, title='Izvirna slika', subplot=221, vmin=0, vmax=255)
  >>> imShowEx(ims, title='Linearno preslikana slika: 0.5*slika + 128', subplot=222, vmin=0, vmax=255)
  >>> pp.subplot(223)
  >>> imHistogram(im, 128, [0, 255], title='')
  >>> pp.subplot(224)
  >>> imHistogram(ims, 128, [0, 255], title='')
  >>>
  '''
  return img.astype(np.float)*slope + intersection  

def imWindow(img, center, width, ls=255):
  '''
  Linearno oknenje.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna referenčna slika.
  center: float
    Središče okna.
  width: float
    Širina okna.
  ls: int
    Maksimalna sivinska vrednosti definicijskega območja oknjene slike 
    (255 za 8-bitno sliko).

  Vrne
  ----
  oimg: numpy.ndarray
    Oknjena slika.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_256x256_uint8.raw', 256, 256)
  >>> imw = imWindow(im, 128, 64, 255)
  >>>  
  >>> imShowEx(im, title='Izvirna slika', subplot=221, vmin=0, vmax=255)
  >>> imShowEx(imw, title='Linearno oknjena slika: center=128, width=64', subplot=222, vmin=0, vmax=255)
  >>> pp.subplot(223)
  >>> imHistogram(im, 32, [0, 255], title='')
  >>> pp.subplot(224)
  >>> imHistogram(imw, 32, [0, 255], title='')
  >>>
  '''
  oimg = np.zeros(img.shape)
  ind1 = img < center - width*0.5
  oimg[ind1] = 0
  ind2 = img > center + width*0.5
  oimg[ind2] = ls
  ind3 = np.logical_not(ind1 | ind2)
  oimg[ind3] = ls/width*(img[ind3] - (center - width*0.5))
  return oimg

def imThreshold(img, threshold, ls=255):
  '''
  Upragovljanje slike.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna referenčna slika.
  threshold : float
    Prag.
  ls: int
    Sivinska vrednost slikovnih elementov, ki so večji ali enaki pragu.      

  Vrne
  ----
  oimg: numpy.ndarray
    Upragovljena slika.

  Primeri uporabe
  ----------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_256x256_uint8.raw', 256, 256)
  >>> imt = imThreshold(im, 100, 255)
  >>>  
  >>> imShowEx(im, title='Izvirna slika', subplot=221, vmin=0, vmax=255)
  >>> imShowEx(imt, title='Upragovljena slika: threshold=100', subplot=222, vmin=0, vmax=255)
  >>> pp.subplot(223)
  >>> imHistogram(im, 256, [0, 255], title='')
  >>> pp.subplot(224)
  >>> imHistogram(imt, 256, [0, 255], title='')
  >>>
  '''
  oimg = np.zeros(img.shape)
  oimg[img < threshold] = 0
  oimg[img >= threshold] = ls
  return oimg

def imMultiscale(img, r, s):
  '''
  Odsekoma linearna sivinska preslikava.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna referenčna slika.
  r: list
    Zaporedje točk, ki določa linearno preslikavo v referenčni sliki.
  s: list
    Zaporedje točk, ki določa linearno preslikavo v preslikani sliki.

  Vrne
  ----
  oimg: numpy.ndarray
    Sliko preslikano z dano odsekoma linearno preslikavo.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_256x256_uint8.raw', 256, 256)
  >>> imms = imMultiscale(im, [0, 255], [255, 0])
  >>>  
  >>> imShowEx(im, title='Izvirna slika', subplot=221, vmin=0, vmax=255)
  >>> imShowEx(imms, title='Odsekoma linearna pres. - invertiranje', subplot=222, vmin=0, vmax=255)
  >>> pp.subplot(223)
  >>> imHistogram(im, 256, [0, 255], title='')
  >>> pp.subplot(224)
  >>> imHistogram(imms, 256, [0, 255], title='')
  >>>
  '''
  oimg = np.zeros(img.shape)
  for i in range(len(r) - 1):
    ind = (img >= r[i]) & (img <= r[i + 1])
    oimg[ind] = (s[i + 1] - s[i])/(r[i + 1] - r[i])* \
      (img[ind] - r[i]) + s[i]
  return oimg

def imGamma(img, gamma, lr=255):
  '''
  Nelinearna Gamma preslikava.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna referenčna slika.
  gamma: float
    Parameter Gama preslikave.
  lr: int
    Največja sivinska vrednost definicijskega območja referenčne slike 
    (255 za 8-bitno sliko).

  Vrne
  ----
  oimg: numpy.ndarray
    Slika preslikana z dano nelinearno Gama preslikavo.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_256x256_uint8.raw', 256, 256)
  >>> img = imGamma(im, 2)
  >>>  
  >>> imShowEx(im, title='Izvirna slika', subplot=221, vmin=0, vmax=255)
  >>> imShowEx(img, title='Gamma preslikava: gamma=2', subplot=222, vmin=0, vmax=255)
  >>> pp.subplot(223)
  >>> imHistogram(im, 128, [0, 255], title='')
  >>> pp.subplot(224)
  >>> imHistogram(img, 128, [0, 255], title='')
  >>>    
  '''
  return lr**(1.0 - gamma)*img.astype(np.float)**gamma

def imProfile2d(img, t1, t2, x=None, y=None, step=1, order=1):
  '''
  Prerez sivinske slike vzdolž daljice t1, t2.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna sivinska slika velikosti (H, W).
  t1: list
    Začetna točka prerezne daljice, podana kot seznam [x1, y1].
  t2: list
    Končna točka prerezne daljice, podana kot seznam [x2, y2].
  x: numpy.ndarray
    Vektor točk v x osi, ki določa definicijsko območje slike. Ko je vrednost
    None, se uporabi privzeti koordinatni sistem np.arange(W), kjer W pomeni
    širino slike, izraženo v slikovnih elementih.
  y: numpy.ndarray
    Vektor točk v y osi, ki določa definicijsko območje slike. Ko je vrednost
    None, se uporabi privzeti koordinatni sistem np.arange(H), kjer H
    pomeni višino slike, izraženo v slikovnih elementih.
  step: float
    Korak vzorčenja vzdolž prerezne daljice.
  order: int
    Red interpolacije sivinskih vrednosti (0 - nearest, 1 - bilinearen).

  Vrne
  ----
  p: numpy.ndarray
    Enakomerno vzorčeni prerez sivinske slike vzdolž daljice t1-t2, ter
  xi: numpy.ndarray
    y koordinate vzorčnih točk prereza.
  yi: numpy.ndarray
    y koordinate vzorčnih točk prereza.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw2d('./poglavje_2/mr_256x256_uint8.raw', 256, 256)
  >>> p, x, y = imProfile2d(im, [0, 0], [255, 255])
  >>>  
  >>> imShowEx(im, title='Slika', subplot=121, vmin=0, vmax=255)
  >>> pp.plot(x, y, '-b')
  >>> pp.subplot(122)
  >>> pp.plot(p)
  >>> pp.title('Prerez')
  >>>
  '''
  if y is None: y = np.arange(img.shape[0])
  if x is None: x = np.arange(img.shape[1])

  t1 = np.asarray(t1, dtype=np.float)
  t2 = np.asarray(t2, dtype=np.float)
  s = t2 - t1
  n = int(np.ceil(1 + np.linalg.norm(s)/step))

  xi = np.linspace(t1[0], t2[0], n)
  yi = np.linspace(t1[1], t2[1], n)
  op = interp.interp2(xi, yi, x, y, img)
  return op, xi, yi

def imProfile3d(img, t1, t2, x=None, y=None, z=None, step=1, order=1):
  '''
  Prerez sivinske slike vzdolž daljice t1, t2.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna sivinska slika velikosti (D, H, W).
  t1: list
    Začetna točka prerezne daljice, podana kot seznam [x1, y1, z1].
  t2: list
    Končna točka prerezne daljice, podana kot seznam [x2, y2, z1].
  x: numpy.ndarray
    Vektor točk v smeri x koordinatne osi, ki določa definicijsko območje
    slike. Ko je vrednost None, se uporabi privzeti koordinatni sistem 
    np.arange(W), kjer W pomeni širino slike, izraženo v slikovnih elementih.
  y: numpy.ndarray
    Vektor točk v smeri y koordinatne osi, ki določa definicijsko območje 
    slike. Ko je vrednost None, se uporabi privzeti koordinatni sistem 
    np.arange(H), kjer H pomeni višino slike, izraženo v slikovnih elementih.
  z: numpy.ndarray
    Vektor točk v smeri z koordinatne osi, ki določa definicijsko območje 
    slike. Ko je vrednost None, se uporabi privzeti koordinatni sistem 
    np.arange(D), kjer D pomeni globino slike, izraženo v slikovnih 
    elementih.
  step: float
    Korak vzorčenja vzdolž prerezne daljice.
  order: int
    Red interpolacije sivinskih vrednosti (0 - nearest, 1 - bilinearen).

  Vrne
  ----
  p: numpy.ndarray
    Enakomerno vzorčeni prerez sivinske slike vzdolž daljice t1-t2, ter
  xi: numpy.ndarray
    y koordinate vzorčnih točk prereza.
  yi: numpy.ndarray
    y koordinate vzorčnih točk prereza.
  zi: numpy.ndarray
    z koordinate vzorčnih točk prereza.

  Primeri uporabe
  ---------------
  >>> im = imLoadRaw3d('./poglavje_2/mr_217x181x181_uint8.raw', 217, 181, 181)
  >>> p, x, y, z = imProfile3d(im, [0, 0, 0], [217, 181, 181])
  >>>
  >>> pp.plot(p)
  >>> pp.title('Diagonalni prerez')
  >>>
'''
  if z is None: z = np.arange(img.shape[0])
  if y is None: y = np.arange(img.shape[1])
  if x is None: x = np.arange(img.shape[2])

  t1 = np.asarray(t1, dtype=np.float)
  t2 = np.asarray(t2, dtype=np.float)
  s = t2 - t1
  n = int(np.ceil(1 + np.linalg.norm(s)/step))

  xi = np.linspace(t1[0], t2[0], n)
  yi = np.linspace(t1[1], t2[1], n)
  zi = np.linspace(t1[2], t2[2], n)
  op = interp.interp3(xi, yi, zi, x, y, z, img)

  return op, xi, yi, zi

#%% Poglavje 6

def imShadingCalibrate(img, dark=None, bright=None, d=0, b=255):
  '''
  Obnova sivinskih vrednosti s kalibracijo.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  dark: numpy.ndarray
    Kalibracijska slika temnega ozadja. Ko je vrednost parametra None se
    izvede kalibracija z multiplikativnim modelom.
  bright: numpy.ndarray
    Kalibracijska slika svetlega ospredja. Ko je vrednost parametra None se
    izvede kalibracija z aditivnim modelom.
  d: float
    Sivinska vrednost ozadja po kalibraciji.
  b: float
    Sivinska vrednost ospredja po kalibraciji.

  Vrne
  ----
  oimg: numpy.ndarray
    Sivinsko obnovljena slika.
  ofa: numpy.ndarray
    Aditivno polje nehomogenosti.
  ofm: numpy.ndarray
    Multiplikativno polje nehomogenosti.

  Primeri uporabe
  ---------------
  >>> from PIL import Image as im
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> dark = np.array(im.open('./vaja5/vaja5_dark.png'))
  >>> bright = np.array(im.open('./vaja5/vaja5_bright.png'))
  >>> image = np.array(im.open('./vaja5/vaja5_image.png'))
  >>>
  >>> imageh, af, mf = imShadingCalibrate(image, dark, bright)
  >>>
  >>> pp.figure()
  >>>  
  >>> pp.subplot(221)
  >>> pp.imshow(image, cmap='gray')
  >>> pp.title('Nehomogena')
  >>>
  >>> pp.subplot(222)
  >>> pp.imshow(imageh, cmap='gray')
  >>> pp.title('Obnovljena')
  >>>
  >>> pp.subplot(223)
  >>> pp.imshow(af, cmap='gray')
  >>> pp.title('Aditivno polje')
  >>>  
  >>> pp.subplot(224)
  >>> pp.imshow(mf, cmap='gray')
  >>> pp.title('Multiplikativno polje')
  >>>
  '''
  img = np.asarray(img, np.float)
  if dark is not None:
    dark = np.asarray(dark, np.float)
  if bright is not None:
    bright = np.asarray(bright, np.float)

  if bright is None:
    ofa = dark - d
    ofm = None
    oimg= img - dark + d

  elif dark is None:
    ofa = None
    ofm = bright/b
    oimg = img*b/bright

  else:
    ofa = (d*dark  - b*bright)/float(b - d)
    ofm = (bright - dark)/float(b - d)
    oimg = (img*(b - d) - b*dark + d*bright)/(bright - dark)

  return oimg, ofa, ofm

def imShadingFilter(img, sigma):
  '''
  Obnova sivinskih vrednosti s filtriranjem. Predpostavljamo
  aditivno polje svetlostne nehomogenosti.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  sigma: float
    Standardna deviacija gaussovega filtra.

  Vrne
  ----
  oimg: numpy.ndarray
    Sivinsko obnovljena slika.
  af: numpy.ndarray
    Aditivno polje nehomogenosti.

  Primeri uporabe
  ---------------
  >>> from PIL import Image as im
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> image = np.array(im.open('./poglavje_6/shading_muscle.png'))
  >>>
  >>> imageh, af = imShadingFilter(image, 50)
  >>>
  >>> pp.figure()
  >>>  
  >>> pp.subplot(131)
  >>> pp.imshow(image, cmap='gray')
  >>> pp.title('Nehomogena')
  >>>
  >>> pp.subplot(132)
  >>> pp.imshow(imageh, cmap='gray')
  >>> pp.title('Obnovljena')
  >>>
  >>> pp.subplot(133)
  >>> pp.imshow(af, cmap='gray')
  >>> pp.title('Aditivno polje')
  >>>  
  '''
  img = img.astype(np.float)
  #af = imGaussFilt2d(logimg, float(sigma), boundary='reflect')
  af = ndim.gaussian_filter(img, sigma, mode='reflect')
  oimg = img - af

  return oimg, af

  
def imShadingHomomorphic(img, sigma):
  '''
  Obnova sivinskih vrednosti s homomorfnim filtriranjem. Predpostavljamo
  multiplikativno polje svetlostne nehomogenosti.

  Parametri
  ---------
  img: numpy.ndarray
    Vhodna slika.
  sigma: float
    Standardna deviacija gaussovega filtra.

  Vrne
  ----
  oimg: numpy.ndarray
    Sivinsko obnovljena slika.
  mf: numpy.ndarray
    Multiplikativno polje nehomogenosti.

  Primeri uporabe
  ---------------
  >>> from PIL import Image as im
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> image = np.array(im.open('./poglavje_6/shading_muscle.png'))
  >>>
  >>> imageh, mf = imShadingHomomorphic(image, 25)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(131)
  >>> pp.imshow(image, cmap='gray')
  >>> pp.title('Nehomogena')
  >>>
  >>> pp.subplot(132)
  >>> pp.imshow(imageh, cmap='gray')
  >>> pp.title('Obnovljena')
  >>>
  >>> pp.subplot(133)
  >>> pp.imshow(mf, cmap='gray')
  >>> pp.title('Aditivno polje')
  >>>
  '''
  logimg = np.log(np.asarray(img, dtype=np.float) + 1.0)
  # logimgf = imGaussFilt2d(logimg, float(sigma), boundary='reflect')
  logimgf = ndim.gaussian_filter(logimg, sigma=sigma, mode='reflect')
  mf = np.exp(logimgf) - 1.0
  oimg = img.astype(np.float)/mf

  return oimg, mf

#%% Poglavje 7

def transformAffine2d(scale=[1.0, 1.0], trans=[0.0, 0.0],
                      rot=0.0, shear=[0.0, 0.0]):
  '''
  Zgradi afino transformacijsko matriko. Vrstni red operacij je sledeč:
  skaliranje, translacija, rotiranje in strig.

  Parametri
  ---------
  scale: list
    Skaliranje v x in y smeri podana kot (sx, sy).
  trans: list
    Translacija/premik v x in y smeri podana kot (tx, ty).
  rot: float
    Kot rotacije v radianih.
  shear: list
    Parametra striga v x in y smeri podana kot (sx, sy).
  
  Vrne
  ----
  T: numpy.ndarray
    Transformacijska matrika Tskaliranja*Ttranslacije*Trotacije*Tstriga.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> T = transformAffine2d(scale=[2, 2], rot=np.deg2rad(45))
  >>> x = np.arange(10)
  >>> y = np.arange(8)
  >>> Y, X = np.meshgrid(y, x, indexing='ij')
  >>> tmp = np.dot(T, np.vstack([X.flat, Y.flat, np.ones([X.size])]))
  >>> Yt, Xt = tmp[0], tmp[1]
  >>> 
  >>> pp.figure()
  >>> pp.plot(X.flat, Y.flat, '.g', label='Izvirne')
  >>> pp.plot(Xt.flat, Yt.flat, '.r', label='Transformirane')
  >>> pp.legend()
  >>>
  '''
  
  fi = rot

  S = np.array([[scale[0], 0, 0],
                [0, scale[1], 0],
                [0, 0, 1.0]], dtype=np.float)

  T = np.array([[1, 0, trans[0]],
                [0, 1, trans[1]],
                [0, 0, 1.0]], dtype=np.float)

  R = np.array([[np.cos(fi), -np.sin(fi), 0],
                [np.sin(fi), np.cos(fi), 0],
                [0, 0, 1.0]], dtype=np.float)

  H = np.array([[1, shear[0], 0],
                [shear[1], 1, 0],
                [0, 0, 1.0]], dtype=np.float)

  return np.dot(H, R).dot(T).dot(S)

def ptTransform2d(t, x, y, inverse=False):
  '''
  Preslika točke (x,y) s podano projektivno preslikavo.

  Parametri
  ---------
  t: numpy.ndarray
    Transformacijska matrika velikosti 3 x 3.
  x: numpy.ndarray
    Podatkovno polje, ki določa x koordinate točk. Velikosti podatkovnih
    polj x in y se morata ujemati.
  y: numpy.ndarray
    Podatkovno polje, ki določa y koordinate točk. Velikosti podatkovnih
    polj x in y se morata ujemati.
  inverse: bool
    Ko je vrednost True, se točke preslikajo z inverzom preslikave t.

  Vrne
  -----
  ox: numpy.ndarray
    X koordinate transformiranih točk. Velikost podatkovnega polja se ujema
    z velikostjo podatkovnega polja x.
  oy: numpy.ndarray
    X koordinate transformiranih točk. Velikost podatkovnega polja se ujema
    z velikostjo podatkovnega polja y.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>>
  >>> T = transformAffine2d(scale=[2, 2], rot=np.deg2rad(45))
  >>> x = np.arange(10)
  >>> y = np.arange(8)
  >>> Y, X = np.meshgrid(y, x, indexing='ij')
  >>> Yt, Xt = ptTransform2d(T, X, Y)
  >>> 
  >>> pp.figure()
  >>> pp.plot(X.flat, Y.flat, '.g', label='Izvirne')
  >>> pp.plot(Xt.flat, Yt.flat, '.r', label='Transformirane')
  >>> pp.legend()
  >>>
  '''
  x = np.asarray(x, dtype=np.float)
  y = np.asarray(y, dtype=np.float)
  t = np.asarray(t)
  R = np.vstack(
    (x.flatten(),
    y.flatten(),
    np.ones([x.size]))
  )
  if inverse:
    Rt = np.linalg.solve(t, R)
  else:
    Rt = np.dot(t, R)

  ox, oy = Rt[0, :]/Rt[-1, :],  Rt[1, :]/Rt[-1, :]
  ox.shape, oy.shape = x.shape, y.shape

  return ox, oy

def imTransform2d(img, t, x=None, y=None, center=(0.0, 0.0), 
                  order=1, expand='same'):
  '''
  Projektivna preslikava slike.
  
  Parametri
  ---------
  img: numpy.ndarray
    Vhodna sivinska slika.
  t : podatkovno polje 3 x 3
    Transformacijska matrika velikosti 3 x 3.
  x: numpy.ndarray
    Vektor točk, ki določa koordinatni sistem vhodne slike v smeri x osi. Ko
    je x enak None, se uporabi privzeto definicijsko območje 
    x = np.arange(img.shape[1]).
  y: numpy.ndarray
    Vektor točk, ki določa koordinatni sistem vhodne slike v smeri y osi. Ko
    je y enak None, se uporabi privzeto definicijsko območje 
    y = np.arange(img.shape[0]).
  center : seznam ali podatkovno polje velikosti 2
    Središče preslikave
  order: int
    Red interpolacije (0 - najbližji sosed, 1 - bilinearna, ...).
  expand: str
    Definicijsko območje preslikane slike: 

        'same' - enako vhodni sliki,

        'full' - enako uniji vhodne in preslikane slike,

        'crop' - enako preslikani sliki.

  Vrne
  ----
  imgt: numpy.ndarray
      Slika preslikana s preslikavo T.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = np.array(im.open('./poglavje_7/preslikave_ct.png'))
  >>> T = transformAffine2d(scale=[1.2, 1.2], rot=np.deg2rad(5))
  >>> st = imTransform2d(s, T)
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(121)
  >>> pp.title('Izvirna slika')
  >>> pp.imshow(s, cmap='gray')  
  >>>
  >>> pp.subplot(122)
  >>> pp.title('Preslikana slika')  
  >>> pp.imshow(st, cmap='gray')
  >>>
  '''
  H, W = img.shape
  if x is None:
    x = np.arange(W) - center[0]
  if y is None:
    y = np.arange(H) - center[1]

  dx, dy = x[1] - x[0], y[1] - y[0]
  
  if expand == 'same':
    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()

  elif expand in ['full', 'crop']:
    # oglišča referenčne slike v preslikani sliki določajo meje razširitve
    etx, ety = ptTransform2d(t,
      [x[0], x[0],  x[-1], x[-1]],
      [y[0], y[-1], y[0], y[-1]])

    xmin, xmax = np.floor(etx).min(), np.floor(etx).max()
    ymin, ymax = np.ceil(ety).min(), np.ceil(ety).max()

    if expand == 'full':
      xmin = min(x.min(), xmin)
      xmax = min(x.max(), xmax)
      ymin = min(y.min(), ymin)
      ymax = min(y.max(), ymax)

  xt = np.arange(xmin, xmax + dx, dx)
  yt = np.arange(ymin, ymax + dy, dy)
  Yt, Xt = np.meshgrid(yt, xt, indexing='ij')
  Xi, Yi = ptTransform2d(t, Xt, Yt, inverse=True)
  
  oimg = interp.interp2(Xi, Yi, x, y, img.astype(np.float))

  return oimg

def imRotate2d(img, angle, order=1, expand='same'):
  '''
  Rotacija slike okoli središča slike. Za opis glej parametre funkcije
  imTransform2d.

  Parametri
  ---------
  angle: float
    Kot rotacije v radianih.
  order, expand :
    Za opis glej parametre funkcije imTransform2d.

  Vrne
  ----
  oimg: numpy.ndarray
    Slika rotirana okoli središča.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = np.array(im.open('./vaja6/vaja6.png'))
  >>> st = imRotate2d(s, angle=np.deg2rad(45))
  >>>
  >>> pp.figure()
  >>>
  >>> pp.subplot(121)
  >>> pp.title('Izvirna slika')
  >>> pp.imshow(s, cmap='gray')  
  >>>
  >>> pp.subplot(122)
  >>> pp.title('Preslikana slika')  
  >>> pp.imshow(st, cmap='gray')
  >>>
  '''
  H, W = img.shape
  x, y = np.arange(W, dtype=np.float), np.arange(H, dtype=np.float)
  x -= x.mean()
  y -= y.mean()
  T = transformAffine2d(rot=angle)

  return imTransform2d(img, T, x, y,
                       expand=expand, order=order)

#%% Poglavje 8

def transformEstimate2d(x, y, u, v, kind='rigid', nr=1, 
                        animate=False):
  '''
  Določi izbrano intrerpolacijsko ali aproksimacjsko preslikavo na
  podlagi podanih korespondečnih točk (x, y) -T-> (u,v).

  Parametri
  ---------
  x, y: numpy.ndarray
    Koordinate korespondečnih točk v referenčni sliki.
  u, v: numpy.ndarray
    Koordinate korespondenčnih točk v zajeti sliki.
  kind: str
    Model preslikave 'rigid', 'affine', 'projective', ali radial.
  nr: int
    Red radialnih distorzij.
  animate: bool
    Ko je vrednost parametra True, se prikazujejo vmesni rezultati 
    optimizacijskega postopka.

  Vrne
  ----
  t: numpy.ndarray
    Transformacijska matrika velikosti 3 x 3.
  tr: numpy.ndarray
    Parametri radialnih distrorzij. Vunkcija vrne vrednost le takrat, ko
    je vrednost parametra kind enaka "radial".
  err: float
    Napaka poravnave med korespondenčnimi točkami.  

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>>
  >>> x, y = np.random.rand(10), np.random.rand(10)
  >>> t = transformAffine2d(scale=[2.0, 0.5], trans=[5.0, 2.0], rot=np.pi/4, shear=[0.2, 0.5])
  >>> u, v = ptTransform2d(t, x, y)
  >>> te, err = transformEstimate2d(x, y, u, v, kind='affine')
  >>> t, te
  >>>
  '''
  x = np.asarray(x).astype(np.float).flatten()
  y = np.asarray(y).astype(np.float).flatten() 
  u = np.asarray(u).astype(np.float).flatten()
  v = np.asarray(v).astype(np.float).flatten()
  t = tr = err = None
  
  if kind == 'rigid':
    xm = x.mean()
    um = u.mean()
    ym = y.mean()
    vm = v.mean()
    yum = np.mean(y*u)
    xvm = np.mean(x*v) 
    xum = np.mean(x*u)
    yvm = np.mean(y*v)
    a = -np.arctan(
      (yum - xvm - ym*um + xm*vm)/
      (xum +  yvm - xm*um - ym*vm))
    tx = um - xm*np.cos(a) + ym*np.sin(a)
    ty = vm - xm*np.sin(a) - ym*np.cos(a)
    t = np.dot(
      transformAffine2d(trans=[tx, ty]),
      transformAffine2d(rot=a))
    tx, ty = ptTransform2d(t, x, y)
    err = (((u - tx)**2 + (v - ty)**2).mean())**0.5
    return t, err

  elif kind == 'affine':
    xxm = np.mean(x*x)
    xym = np.mean(x*y)
    yym = np.mean(y*y)
    xm = x.mean()
    ym = y.mean() 
    uxm = np.mean(x*u)
    uym = np.mean(u*y) 
    vxm = np.mean(v*x)
    vym= np.mean(y*v)
    um = u.mean()
    vm = v.mean()
    t = np.zeros([3,3])
    t[-1, -1] = 1.0
    tvec = np.linalg.solve(
      np.array(
        [[xxm, xym, xm, 0, 0, 0],
        [xym, yym, ym, 0, 0, 0],
        [xm, ym, 1, 0, 0, 0],
        [0, 0, 0, xxm, xym, xm],
        [0, 0, 0, xym, yym, ym],
        [0, 0, 0, xm, ym, 1]]
      ),
      np.array([uxm, uym, um, vxm, vym, vm]))
    t[0] = tvec[:3]
    t[1] = tvec[3:]
    tx, ty = ptTransform2d(t, x, y)
    err = (((u - tx)**2 + (v - ty)**2).mean())**0.5
    return t, err

  elif kind == 'projective':
    if animate:
      pp.figure()
    t0, err = transformEstimate2d(x, y, u, v, 'affine')
    t0 = t0.flatten()[:-1]
    topt = opt.fmin(
      lambda t: kProjective2d(t, x, y, u, v, animate), t0)
    t = np.ones([9], 'float')
    t[:8] = topt
    t.shape = [3, 3]
    err = kProjective2d(topt, x, y, u, v, False)
    return t, err

  elif kind == 'radial':
    if animate:
      pp.figure()
    Tproj = np.zeros([3, 3])
    xc = x.mean()
    yc = y.mean()
    tr0 = np.zeros([nr + 2])
    tr0[0] = xc
    tr0[1] = yc
    tropt = opt.fmin(
      lambda tr: kProjectiveRadial2d(tr, x, y, u, v, Tproj, animate), 
      tr0)
    tr = tropt
    err = kProjectiveRadial2d(tropt, x, y, u, v, Tproj, False)
    return Tproj, tr, err

  else:
    raise ValueError(
      'Vrednost parametra "kind" je lahko '
      '"rigid", "affine", "projective" ali "radial"!')

  return t, tr, err

def kProjective2d(t, x, y, u, v, animate):
  '''
  Kriterijska funkcija za določevanje parametrov projektivne preslikave na
  podlagi podanega seta korespondenčnih točk (x,y) ter (u,v).

  Parametri
  ---------
  t: numpy.ndarray
    Trenutna vrednost parametrov projektivne preslikave, ki so zapisani v
    vektorju (po vrsticah).
  x, y: numpy.ndarray
    Koordinate korespondečnih točk v referenčni sliki.
  u, v: numpy.ndarray
    Koordinate korespondenčnih točk v zajeti sliki.
  animate: bool
    Ko je vrednost parametra True, se prikazujejo vmesni rezultati 
    optimizacijskega postopka.

  Vrne
  ----
  err: float
    Vrne koren srednje kvadratične razdalje med preslikanimi referenčnimi 
    točkami in točkami v zajeti sliki.
  '''
  T = np.zeros([9])
  T[:8] = t
  T[-1] = 1.0
  T.shape = (3,3)
  ue, ve = ptTransform2d(T, x, y)
  if animate:
    pp.cla()
    pp.plot(u, v, 'xr')
    pp.plot(ue, ve, 'xb')
    pp.draw()
    pp.pause(0.01)

  return (((u - ue)**2 + (v - ve)**2).mean())**0.5

def ptTransformRadial2d(tr, x, y):
  '''
  Preslika točke s podano radialno preslikavo.

  Parametri
  ---------
  tr: numpy.ndarray
    Radialna preslikava podana kot [xc, yc, K]
  x, y: numpy.ndarray
    Koordinate točk.

  Vrne
  ----
  xt: numpy.ndarray
    Preslikane točke x.
  yt: numpy.ndarray
    Preslikane točke y.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>>
  >>> x = y = np.arange(5)
  >>> Y, X = np.meshgrid(y, x, indexing='ij')
  >>> U, V = ptTransformRadial2d([2.0, 2.0, 0.1], X, Y)
  >>>
  >>> pp.figure()
  >>> pp.plot(X.flat, Y.flat, 'xr', label='referenčne')
  >>> pp.plot(U.flat, V.flat, 'og', label='preslikane')
  '''
  x = np.asarray(x)
  y = np.asarray(y)
  
  tr = np.asarray(tr, 'float')

  xc = tr[0]
  yc = tr[1]
  K = tr[2:]

  rr = (x - xc)**2 + (y - yc)**2
  sr = 1
  for i in range(len(K)):
    sr += K[i]*(rr**(i + 1.0))

  xt = xc + (x - xc)*sr
  yt = yc + (y - yc)*sr

  return xt, yt

def kProjectiveRadial2d(t, x, y, u, v, Tproj, animate):
  '''
  Kriterijska funkcija za določevanje parametrov projektivne in radialne 
  preslikave na podlagi podanega seta korespondenčnih točk (x,y) ter (u,v).

  Parametri
  ---------
  t: numpy.ndarray
    Trenutna vrednost parametrov radialne preslikave, ki so 
    zapisani v vektorju [xcenter, ycenter, K].
  x, y: numpy.ndarray
    Koordinate korespondečnih točk v referenčni sliki.
  u, v: numpy.ndarray
    Koordinate korespondenčnih točk v zajeti sliki.
  Tproj: numpy.ndarray
    Matrika projektivne preslikave velikosti 3 x 3. Se posodobi ob vsakem klicu.
  animate: bool
    Ko je vrednost parametra True, se prikazujejo vmesni rezultati 
    optimizacijskega postopka.

  Vrne
  ----
  err: float
    Vrne koren srednje kvadratične razdalje med preslikanimi referenčnimi 
    točkami in točkami v zajeti sliki.

  Opombe
  ------
    V globalni spremenljivko Tprojective_ se nahajajo parametri projektivne 
    preslikave, ki preslika z radialno preslikavo preslikane referenčne točke v
    prostor zajete slike.
  '''
  # najprej izvedemo radialno preslikavo  
  xr, yr = ptTransformRadial2d(t, x, y)

  # in se aproksimacijska projektivna preslikava
  t, err = transformEstimate2d(xr, yr, u, v, 
    'projective', False)
  [ue, ve] = ptTransform2d(t, xr, yr)
  Tproj[:, :] = t

  if animate:
    pp.cla()
    pp.plot(u, v, 'xr')
    pp.plot(ue, ve, 'xb')
    pp.draw()
    pp.pause(0.01)

  return (((u - ue)**2 + (v - ve)**2).mean())**0.5


#%% Poglavje 9

def imSM(imageA, imageB, sm, nb=16, span=(0,255)):
  '''
  Izračuna vrednost izbrane mere podobnosti za podani sliki.

  Partametri
  ----------
  imageA, imageB: numpy.ndarray
    Vhodni sliki zapisani v 8-bitnem nepredznačenem
  sm: str
    Izbrana mera podobnosti 'mse', 'mae', 'cc' ali 'mi'.
  nb: int
    Število razredov histograma pri izračunu mere podobnosti 'mi'.
  span: list, tuple
    Razpon sivinskih vrednosti kot [min, max]

  Vrne
  ----
  mi: float
    Izračunano vrednost mere podobnosti.
  '''
  imageA = np.asarray(imageA, dtype=np.float)
  imageB = np.asarray(imageB, dtype=np.float)
  N = imageA.size
  nb, span = int(nb), np.asarray(span, dtype=np.float)
  sm = str(sm).lower()
  if sm == 'mse':
    of = ((imageA - imageB)**2).mean()

  elif sm == 'mae':
    of = (np.abs(imageA - imageB)).mean()

  elif sm == 'cc':
    imageAm = imageA.mean()
    imageBm = imageB.mean()
    of = ((imageA - imageAm)*(imageB - imageBm)).sum()/ \
      np.sqrt(((imageA-imageAm)**2).sum()*((imageB - imageBm)**2).sum())

  elif sm == 'mi':
    d = span[1] - span[0]
    imageAd = np.round((imageA - span[0])/(d/nb))
    imageBd = np.round((imageB - span[0])/(d/nb))
    pa, pb, pab = np.zeros([nb]), np.zeros([nb]), np.zeros([nb, nb])
    for i in range(nb):
      pa[i] = np.count_nonzero(imageAd == i)
      pb[i] = np.count_nonzero(imageBd == i)

    pa *= 1.0/N
    pb *= 1.0/N

    for i in range(nb):
      for j in range(nb):
        pab[i, j] = np.count_nonzero((imageAd == i) & (imageBd == j))

    pab *= 1.0/N

    pa[pa == 0] = 1.0
    pb[pb == 0] = 1.0
    pab[pab == 0] = 1.0
    Ha, Hb = -(pa*np.log(pa)).sum(), -(pb*np.log(pb)).sum()
    Hab = -(pab*np.log(pab)).sum()
    of = Ha + Hb - Hab
  else:
    raise ValueError('Vrednost parametra sm je lahko '
                     '"mse", "mae", "cc" ali "m1"!.')

  return of

def imRigidRegister(imageA, imageB, sm, x0=[0.0, 0.0, 0.0], 
                    animate=False):
  '''
  Togo poravna lebdečo sliko imageB na referenčno sliko imageA.

  Parametri
  ---------
  imageA: numpy.ndarray
    Referenčna slika.
  imageB: numpy.ndarray
    Lebdeča slika.
  sm: str
    Izbrana mera podobnosti.
  x0: list, tuple
    Začetni približek preslikave podan kot [tx, ty, rot].
  animate: bool
    Ko je vrednost parametra True, se izrisujejo vmesni rezultati poravnave.

  Vrne
  ----
  x: numpy.ndarray
    Vektor optimalnih vrednosti parametrov preslikave.
  t: numpy.ndarray
    Pripadajoča transformacijska matrika velikosti 3 x 3.
  f: float
    Pripadajoča vrednost mere podobnosti.
  n: int
    Število izračunov mere podobnosti.
  '''
  # pripravimo vhodne podatke
  imageA = np.asarray(imageA, dtype=np.float)
  imageB = np.asarray(imageB, dtype=np.float)
  x0 = np.asarray(x0, dtype=np.float)
  sm = str(sm).lower()

  # ponastavimo vrednost števca klicev kriterijske funkcije
  Nreg = np.array(0)

  # pripravimo okno za izrisovanje vmesnih rezultatov
  if animate:
    pp.figure()

  # začnemo poravnavo z optimizaciojo mere podobnosti
  ox = opt.fmin(
    lambda x: kFunRigidReg(x, imageA, imageB, sm, animate, Nreg), x0)
  # izračunamo vrednost kriterijske funkcije  ob zaključku poravnave
  of = kFunRigidReg(ox, imageA, imageB, sm, False)
  # ustvarimo transformacijsko matriko toge preslikave
  ot = transformAffine2d(trans=ox[:2], rot=ox[2])

  return ox, ot, of, int(Nreg)

def kFunRigidReg(x, img1, img2, sm, animate, Nreg=0):
  '''
  Kriterijska funkcija toge poravnave.

  Parametri
  ---------
  img1: numpy.ndarray
    Referenčna slika.
  img2: numpy.ndarray
    Lebdeča slika.
  sm: str
    Izbrana mera podobnosti.
  N: numpy.ndarray
    Število klicev kriterijske funkcije kot podatkovno polje velikosti 1.
  animate : bool
    Ko je vrednost parametra True, se izrisujejo vmesni rezultati poravnave.

  Vrne
  ----
  err: float
    Vrednost izbrane mere podobnosti.
  '''
  Nreg += 1
  # ustvarim matriko toge preslikave in z njo preslikamo sliko
  T = transformAffine2d(trans=x[:2], rot=x[2])
  img2t = imTransform2d(img2, T, expand='same')
  # izračunamo vrednost mere podobnosti med lebdečo in referenčno sliko
  f = imSM(img1, img2t, sm)
  if sm == 'cc':
    f = 1.0 - f

  elif sm == 'mi':
    f = -f

  # po potrebi prikažemo stanje poravnave
  if animate and (Nreg % 10) == 0:
    pp.imshow(128.0 + img2t - img1, cmap='gray', vmin=0, vmax=255)
    pp.title('Iteracija {}, vrednost kriterijske funkcije {}'.\
      format(Nreg, f))
    pp.draw()
    pp.pause(0.01)

  return f

#%% Poglavje 10

def imParallelBeamProject2d(fi, step, image, x, y,
                            d, n, pixelSize, kind='sum'):
  '''
  Izračuna pravokotno projekcijo slike.

  Parametri
  ---------
  fi: float, numpy.ndarray
    Koti rotacije pri katerih se izračuna pravokotna projekcija.
  step: float
    Korak vzorčenja vzdolž vzorčnih premic.
  image: numpy.ndarray
    Vhodna 2D slika.
  x, y: numpy.ndarray
    Vektorja, ki napenjata koordinatni sistem slike.
  d: float
    Razdalja med linijskim tipalom in izvorom.
  n: int
    Število slikovnih elementov linijskega tipala.
  pixelSize: float
    Velikost slikovnega elementa linijskega tipala.
  kind: str
    Tip projekcije 'sum', 'mean', 'max'.

  Vrne
  ----
  P: numpy.ndarray
    Dvorazsežno podatkovno polje, kjer so projekcije, ki pripadajo 
    posameznim kotom shranjene v vrsticah matrike.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = np.array(im.open('./poglavje_10/proj2d_test.png'))
  >>> H, W = s.shape
  >>> x = np.arange(-2*(W-1)*0.5, 2*(W-1)*0.5 + 1, 2)
  >>> y = np.arange(-2*(H-1)*0.5, 2*(H-1)*0.5 + 1, 2)
  >>> fi = np.arange(0, 2*np.pi, np.pi/180)
  >>>
  >>> p = imParallelBeamProject2d(fi, 2, s, x, y, 800, 512, 1.0)
  >>>
  >>> pp.figure()
  >>> imShowEx(s, title='Testna slika', subplot=121, extent=[x[0], x[-1], y[0], y[-1]])
  >>> imShowEx(p, title='Sinogram', subplot=122)
  >>>
  '''
  # pripravimo vhodne podatke
  kind = str(kind).lower()
  if kind == 'sum':
    fun = np.sum
  elif kind == 'max':
    fun = np.max
  elif kind == 'mean':
    fun = np.mean
  else:
    raise ValueError(
      'Vrednost parametra "kind" je lahko "sum" ali "max"!')

  image = np.asarray(image, dtype=np.float)
  x = np.asarray(x, dtype=np.float)
  y = np.asarray(y, dtype=np.float)
  if isinstance(fi, float) or isinstance(fi, int): 
    fi = [fi]
  fi = np.asarray(fi, dtype=np.float)
  pixelSize = float(pixelSize)
  step = float(step)
  d = float(d)

  # položaj vira
  ys = pixelSize*np.arange(n, dtype=np.float)
  ys -= ys.mean()
  xs = np.tile(d*0.5, ys.shape)
  # položaji tipal
  # yd = ys
  # xd = -d*0.5

  # vzorčne točke vzdolž projekcijskih daljic
  t = np.arange(0, d + step, step)

  # ustvari projekcijske žarke (vsak v svoji vrstici)
  Xsamples = np.zeros([n, t.size])
  Ysamples = np.zeros([n, t.size])
  for i in range(n):
    Xsamples[i,:] = xs[i] + t*(-1.0)
    Ysamples[i,:] = ys[i] + t*(0.0)
  # vzorčne točke shranimo v dvovrstično matriko - x v prvo, y v drugo vrstico
  pts = np.vstack([Xsamples.flatten(), Ysamples.flatten()])

  # ustvarimo podatkovno polje projekcij/sinograma
  P = np.zeros([fi.size, n])
  for i in range(fi.size):
    # rotiramo vzorčne točke za podani kot rotacije
    tmp = np.dot(np.array([
      [np.cos(fi[i]), -np.sin(fi[i])],
      [np.sin(fi[i]), np.cos(fi[i])]]), pts)
    # interpolacija sivinskih vrednosti v vzorčnih točkah
    proji = interp.interp2(tmp[0], tmp[1], x, y, image)
    # preoblikuj podatke, tako da so žarki v vrsticah
    proji.shape = Xsamples.shape
    # izvedemo zahtevano operacijo vzdolž projekcijskih daljic
    P[i] = fun(proji, 1)

  return P

def imFanBeamProject2d(fi, step, image, x, y,
                       d, n, pixelSize, kind='sum'):
  '''
  Izračuna stožčasto projekcijo slike.
  
  Parametri
  ---------
  fi: float, numpy.ndarray
    Koti rotacije pri katerih se izračuna stožčasta projekcija.
  step: float
    Korak vzorčenja vzdolž vzorčnih premic.
  image: numpy.ndarray
    Vhodna 2D slika.
  x, y: numpy.ndarray
    Vektorja, ki napenjata koordinatni sistem slike.
  d: float
    Razdalja med linijskim tipalom in izvorom.
  n: int
    Število slikovnih elementov linijskega tipala.      
  pixelSize: float
    Velikost slikovnega elementa linijskega tipala.
  kind: str
    Tip projekcije 'sum', 'mean', 'max'.

  Vrne
  ----
  P: numpy.ndarray
    Dvorazsežno podatkovno polje, kjer so projekcije, ki pripadajo 
    posameznim kotom shranjene v vrsticah matrike.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = np.array(im.open('./poglavje_10/proj2d_test.png'))
  >>> H, W = s.shape
  >>> x = np.arange(-2*(W-1)*0.5, 2*(W-1)*0.5 + 1, 2)
  >>> y = np.arange(-2*(H-1)*0.5, 2*(H-1)*0.5 + 1, 2)
  >>> fi = np.arange(0, 2*np.pi, np.pi/180)
  >>>
  >>> p = imFanBeamProject2d(fi, 2, s, x, y, 800, 512, 2.0)
  >>>
  >>> pp.figure()
  >>> imShowEx(s, title='Testna slika', subplot=121, extent=[x[0], x[-1], y[0], y[-1]])
  >>> imShowEx(p, title='Sinogram', subplot=122)
  >>>
  '''
  kind = str(kind).lower()
  if kind == 'sum':
    fun = np.sum
  elif kind == 'max':
    fun = np.max
  elif kind == 'mean':
    fun = np.mean
  else:
    raise ValueError(
      'Vrednost parametra "kind" je lahko "sum" ali "max"!')

  # pripravi sliko, koordinatno mrežo točk in kote
  image = np.asarray(image, dtype=np.float)
  x = np.asarray(x, dtype=np.float)
  y = np.asarray(y, dtype=np.float)
  if isinstance(fi, float) or isinstance(fi, int): 
    fi = [fi]
  fi = np.asarray(fi, dtype=np.float)
  pixelSize = float(pixelSize)
  step = float(step)
  d = float(d)

  # položaj vira
  ys = 0.0
  xs = d*0.5
  # položaji tipal
  yd = pixelSize*np.arange(n, dtype=np.float)
  yd -= yd.mean()
  xd = -d*0.5

  # dolžina najdaljšega žarka
  dmax = np.sqrt(yd[0]**2 + d**2)
  # točke na projekcijski premici
  t = np.arange(0, dmax + step, step)

  # ustvari projekcijske žarke (vsak v svoji vrstici)
  Xsamples = np.zeros([n, t.size])
  Ysamples = np.zeros([n, t.size])
  for i in range(n):
    s = np.array([xd - xs, yd[i] - ys])
    s /= np.linalg.norm(s)
    Xsamples[i,:] = xs + t*s[0]
    Ysamples[i,:] = ys + t*s[1]

  # podatkovno polje sinograma
  P = np.zeros([fi.size, n])
  # interpolacijske točke žarkov
  pts = np.vstack([Xsamples.flatten(), Ysamples.flatten()])
  for i in range(fi.size):
    # rotiraj interpolacijske točke
    tmp = np.dot(np.array([
      [np.cos(fi[i]), -np.sin(fi[i])],
      [np.sin(fi[i]), np.cos(fi[i])]]), pts)
    # interpoliraj
    proji = interp.interp2(tmp[0], tmp[1], x, y, image)
    # preoblikuj podatke, tako da so žarki v vrsticah
    proji.shape = Xsamples.shape
    P[i] = fun(proji, 1)

  return P

#%% Poglavje 11

def imParallelBeamProject3D(
    fi, step, image, x, y, z,
    d, ny, nz, pixelSize, kind='sum'):
  '''
  Izračuna pravokotno projekcijo slike.

  Parametri
  ---------
  fi: float, numpy.ndarray
    Koti rotacije pri katerih se izračuna pravokotna projekcija.
  step: float
    Korak vzorčenja vzdolž vzorčnih premic.
  image: numpy.ndarray
    Vhodna 3D slika.
  x, y, z: numpy.ndarray
    Vektorji, ki napenjajo koordinatni sistem slike.
  d: float
    Razdalja med linijskim tipalom in izvorom.
  ny, nz: int
    Število slikovnih elementov tipala v smeri y in z koordinatnih osi.
  pixelSize : float
    Velikost slikovnega elementa linijskega tipala.
  kind: str
      Tip projekcije 'sum', 'mean', 'max'.

  Vrne
  ----
  P: numpy.ndarray
    Trirazsežno podatkovno polje, kjer so projekcije, ki pripadajo 
    posameznim kotom, shranjene kot [z, y, fi].

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = imLoadRaw3d('./poglavje_11/ct_287x165x194_uint8.raw', 287, 165, 194, np.uint8)
  >>> D, H, W = s.shape
  >>> dz, dy, dx = 4, 2, 2
  >>> x = np.arange(-dx*(W-1)*0.5, dx*(W-1)*0.5 + dx*0.5, dx)
  >>> y = np.arange(-dy*(H-1)*0.5, dy*(H-1)*0.5 + dy*0.5, dy)
  >>> z = np.arange(-dz*(D-1)*0.5, dz*(D-1)*0.5 + dz*0.5, dz)
  >>> fi = np.arange(0, 2*np.pi, np.pi/90)
  >>>
  >>> p = imParallelBeamProject3D(fi, 4.0, s, x, y, z, 800.0, 128, 128, 10.0, 'max')
  >>>
  >>> pp.figure()
  >>> frame = pp.imshow(p[:, :, 0], cmap='gray')
  >>> for i in range(fi.size):
  >>>    frame.set_array(p[:, :, i])
  >>>    pp.draw()
  >>>    pp.pause(.2)
  >>>
  '''

  kind = str(kind).lower()
  if kind == 'sum':
    fun = np.sum
  elif kind == 'max':
    fun = np.max
  else:
    raise ValueError('Vrednost parametera "kind" je lahko '
                     '"max" ali "sum"!')

  # pripravi sliko, koordinatno mrežo točk in kote
  im = np.asarray(image, dtype='float')
  x = np.asarray(x, dtype='float')
  y = np.asarray(y, dtype='float')
  z = np.asarray(z, dtype='float')
  if isinstance(fi, float): fi = [fi]
  fi = np.asarray(fi).astype('float')
  
  # vektorja, ki napenjata pozicije tipal
  yd = pixelSize*(np.arange(ny))
  yd -= yd.mean()
  zd = pixelSize*np.arange(nz)
  zd -= zd.mean()

  # pozicije tipal
  Zd, Yd = np.meshgrid(zd, yd, indexing='ij')
  # Xd = -d*0.5

  # pozicije izvorov
  Xs = d*0.5
  Ys, Zs = Yd, Zd

  # korak vzorčenja
  dmax = d
  t = np.arange(0, dmax, step)
  n = t.size

  # ustvarimo daljice vzdolž katerih vzorčimo 3D sliko
  Xsamples = np.zeros([nz, ny, n])
  Ysamples = np.zeros_like(Xsamples)
  Zsamples = np.zeros_like(Xsamples)
  # smerni vektor vzorčnih daljic s = [-1, 0, 0]
  for i in range(nz):
    for j in range(ny):
      Xsamples[i,j,:] = Xs + (-1)*t
      Ysamples[i,j,:] = Ys[i,j] + (0)*t
      Zsamples[i,j,:] = Zs[i,j] + (0)*t

  # izračunamo sivine vzdolž točk projekcijskih daljic (tretja razsežnost)
  P = np.zeros([nz, ny, fi.size])
  tmppts = np.vstack([Xsamples.flatten(), 
    Ysamples.flatten(), Zsamples.flatten()])
  for i in range(fi.size):
    # rotiraj interpolacijske točke 
    tmp = np.dot(
      np.array([[np.cos(fi[i]), -np.sin(fi[i]), 0.0],
                [np.sin(fi[i]), np.cos(fi[i]),  0.0],
                [0,       0.0,      1.0]]), tmppts)
    # intepolacija
    proji = interp.interp3(tmp[0], tmp[1], tmp[2], x, y, z, im)
    proji.shape = Xsamples.shape
    # integriraj (ali max) vzdolž daljice
    P[:,:,i] = fun(proji, 2)

  return P

def imFanBeamProject3D(
    fi, step, image, x, y, z,
    d, ny, nz, pixelSize, kind='max'):
  '''
  Izračuna stožčasto projekcijo slike.

  Parametri
  ---------
  fi: float, numpy.ndarray
    Koti rotacije pri katerih se izračuna stožčasta projekcija.
  step: float
    Korak vzorčenja vzdolž vzorčnih premic.
  image : podatkovno polje
    Vhodna 3D slika.
  x, y, z: numpy.ndarray
    Vektorji, ki napenjajo koordinatni sistem slike.
  d: float
    Razdalja med linijskim tipalom in izvorom.
  ny, nz: int
    Število slikovnih elementov tipala v smeri y in z koordinatnih osi.
  pixelSize: float
    Velikost slikovnega elementa linijskega tipala.
  kind: str
    Tip projekcije 'sum', 'mean', 'max'.

 Vrne
  ----
  P: numpy.ndarray
    Trirazsežno podatkovno polje, kjer so projekcije, ki pripadajo 
    posameznim kotom, shranjene kot [z, y, fi].

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = imLoadRaw3d('./poglavje_11/ct_287x165x194_uint8.raw', 287, 165, 194, np.uint8)
  >>> D, H, W = s.shape
  >>> dz, dy, dx = 4, 2, 2
  >>> x = np.arange(-dx*(W-1)*0.5, dx*(W-1)*0.5 + dx*0.5, dx)
  >>> y = np.arange(-dy*(H-1)*0.5, dy*(H-1)*0.5 + dy*0.5, dy)
  >>> z = np.arange(-dz*(D-1)*0.5, dz*(D-1)*0.5 + dz*0.5, dz)
  >>> fi = np.arange(0, 2*np.pi, np.pi/90)
  >>>
  >>> p = imFanBeamProject3D(fi, 4.0, s, x, y, z, 800.0, 128, 128, 20.0, 'max')
  >>>
  >>> pp.figure()
  >>> frame = pp.imshow(p[:, :, 0], cmap='gray')
  >>> for i in range(fi.size):
  >>>    frame.set_array(p[:, :, i])
  >>>    pp.draw()
  >>>    pp.pause(.2)
  >>>
  '''

  kind = str(kind).lower()
  if kind == 'sum':
    fun = np.sum
  elif kind == 'max':
    fun = np.max
  else:
    raise ValueError('Vrednost parametera "kind" je lahko '
                     '"max" ali "sum"!')

  # pripravi sliko, koordinatno mrežo točk in kote
  im = np.asarray(image, dtype='float')
  x = np.asarray(x, dtype='float')
  y = np.asarray(y, dtype='float')
  z = np.asarray(z, dtype='float')
  if isinstance(fi, float):
    fi = [fi]
  fi = np.asarray(fi).astype('float')

  # vektorja, ki napenjata pozicije tipal
  yd = pixelSize*(np.arange(ny))
  yd -= yd.mean()
  zd = pixelSize*np.arange(nz)
  zd -= zd.mean()

  # pozicije tipal
  Zd, Yd = np.meshgrid(zd, yd, indexing='ij')
  Xd = -d*0.5

  # pozicije izvorov
  Xs = d*0.5
  Ys, Zs = 0.0, 0.0

  # korak vzorčenja
  dmax = d
  t = np.arange(0, dmax, step)
  n = t.size

  # ustvarimo daljice vzdolž katerih vzorčimo 3D sliko
  Xsamples = np.zeros([nz, ny, n])
  Ysamples = np.zeros_like(Xsamples)
  Zsamples = np.zeros_like(Xsamples)
  # smerni vektor vzorčnih daljic s = [-1, 0, 0]
  for i in range(nz):
    for j in range(ny):
      # smerni vektor žarka
      s = np.array([Xd - Xs, Yd[i,j] - Ys, Zd[i,j] - Zs], 
        dtype='float')
      s /= np.linalg.norm(s)
      Xsamples[i,j,:] = Xs + s[0]*t
      Ysamples[i,j,:] = Ys + s[1]*t
      Zsamples[i,j,:] = Zs + s[2]*t
    
  # izračunamo sivine vzdolž točk projekcijskih daljic (tretja razsežnost)
  P = np.zeros([nz, ny, fi.size])
  tmppts = np.vstack([Xsamples.flatten(), 
    Ysamples.flatten(), Zsamples.flatten()])
  for i in range(fi.size):    # rotiraj interpolacijske točke 
    tmp = np.dot(
      np.array([[np.cos(fi[i]), -np.sin(fi[i]), 0.0],
                [np.sin(fi[i]), np.cos(fi[i]),  0.0],
                [0,       0.0,      1.0]]), tmppts)
    # intepolacija
    proji = interp.interp3(tmp[0], tmp[1], tmp[2], x, y, z, im)
    proji.shape = Xsamples.shape
    # integriraj (ali max) vzdolž daljice
    P[:,:,i] = fun(proji, 2)

  return P

#%% Poglavje 12

def filterSinogram(sinogram, filt='hamming'):
  '''
  Filtrira sinogram z željenim visokoprepustnim sitom.

  Parametri
  ---------
  sinogram: numpy.ndarray
    Sinogram slike s posameznimi projekcijami v vrsticah matrike.
  filt: str, numpy.ndarray
    Tip visokoprepustnega filtra podan kot znakovni niz 
    ("ramp", "shepp-logan", "cosine", "hamming" ali "hann") ali vektor
    konvolucijskega jedra.

  Vrne
  ----
  oSinogram: numpy.ndarray
    Filtrirani songram enake velikosti kot vhodni sinogram.
  '''
  if isinstance(filt, str):
    filt = hpfilter.create(filt)
  oSinogram = np.zeros(sinogram.shape, dtype=np.float)
  for i in range(sinogram.shape[0]):
    oSinogram[i] = convolve(sinogram[i], filt, 'same')

  return oSinogram

def imParallelBeamBackproject2D(fi, sinogram, x, y,
                                d, pixelSize):
  '''
  Izračuna pravokotno povratno projelkcijo iz podanega sinograma.

  Parametri
  ---------
  fi: float, numpy.ndarray
    Koti rotacij pri katerih so podane projekcije sinograma.
  sinogram : 2D podatkovno polje
    Sinogram slike s posameznimi projekcijami v vrsticah matrike.
  x, y: vektor
    Vektorja, ki napenjata koordinatni sistem rekonstruirane slike.
  d: float
    Razdalja med linijskim tipalom in izvorom.
  pixelSize: float
    Velikost slikovnega elementa linijskega tipala.

  Vrne
  ----
  img: numpy.ndarray
    Rekonstruirana slika.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = np.array(im.open('./poglavje_10/proj2d_test.png'))
  >>> H, W = s.shape
  >>> x = np.arange(-2*(W-1)*0.5, 2*(W-1)*0.5 + 1, 2)
  >>> y = np.arange(-2*(H-1)*0.5, 2*(H-1)*0.5 + 1, 2)
  >>> fi = np.arange(0, 2*np.pi, np.pi/180)
  >>>
  >>> p = imParallelBeamProject2d(fi, 2, s, x, y, 800, 512, 1.0)
  >>> p_f = filterSinogram(p, 'hamming')
  >>>
  >>> sr = imParallelBeamBackproject2D(fi, p, x, y, 800, 1.0)
  >>> sr_f = imParallelBeamBackproject2D(fi, p_f, x, y, 800, 1.0)
  >>>
  >>> pp.figure()
  >>> imShowEx(s, title='Testna slika', subplot=221, extent=[x[0], x[-1], y[0], y[-1]])
  >>> imShowEx(p, title='Sinogram', subplot=222)
  >>> imShowEx(sr, title='Rekonstrukcija brez filtriranjem', subplot=223)
  >>> imShowEx(sr_f, title='Rekonstrukcija s filtriranjem', subplot=224)
  >>>
  '''
  # priprava podatkov
  if isinstance(fi, float) or isinstance(fi, int): 
    fi = [fi]  
  fi = np.asarray(fi, dtype=np.float)
  sinogram = np.asarray(sinogram, dtype=np.float)
  x = np.asarray(x, dtype=np.float)
  y = np.asarray(y, dtype=np.float)
  d = float(d)
  pixelSize = float(pixelSize)

  # število slikovnih elementov linijskega tipala
  n = sinogram.shape[1]

  # koordinate slikovnih elementov rekonstruirane slike
  [Yg, Xg] = np.meshgrid(y, x, indexing='ij')
  # koordinate razporedimo v dvovrstično matriko - x v prvo, y v drugo vrstico
  pts = np.vstack([Xg.flatten(), Yg.flatten()])

  # akumulator povratne projekcije
  oBP = np.zeros(Yg.shape)
  for i in range(fi.size):
    # rotacija slikovnih elementovb v obratni smeri 
    #  kot bi rotirali tipalo-izvor
    Tr = np.array([[np.cos(-fi[i]), -np.sin(-fi[i])],
                   [np.sin(-fi[i]),  np.cos(-fi[i])]])
    ptsr = np.dot(Tr, pts)   
    # y koordinate projiciranih slikovnih elementov na tipalu izražene z
    #  naslovom tipala
    yt = ptsr[1,:]/pixelSize + n*0.5
    # poiščemo projekcije z veljavnimi naslovi
    validinds = (yt >= 0) & (yt <= n - 1)
    yt = yt[validinds]

    # z linearno interpolacijo določimo vrednosti v projiciranih točkah
    ind = np.floor(yt)
    delta = yt - ind
    intind = ind.astype('int')
    vt = sinogram[i, intind]*(1.0 - delta) +  \
      sinogram[i, np.minimum(intind + 1, n - 1)]*delta
    validinds.shape = Xg.shape

    # prištejemo izračunane vrednosti akumulatorju povratne projekcije
    oBP[validinds] += vt

  return oBP

def imFanBeamBackproject2D(fi, sinogram, x, y,
                           d, pixelSize):
  '''
  Izračuna stožčasto povratno projelkcijo iz podanega sinograma.

  Parametri
  ---------
  fi: float, list, numpy.ndarray
    Koti rotacij pri katerih so podane projekcije sinograma.
  sinogram: numpy.ndarray
    Sinogram slike s posameznimi projekcijami v vrsticah matrike.
  x, y: numpy.ndarray
    Vektorja, ki napenjata koordinatni sistem rekonstruirane slike.
  d: float
    Razdalja med linijskim tipalom in izvorom.
  pixelSize: float
    Velikost slikovnega elementa linijskega tipala.

  Vrne
  ----
  img: numpy.ndarray
    Rekonstruirana slika.

  Primeri uporabe
  ---------------
  >>> import numpy as np
  >>> from matplotlib import pyplot as pp
  >>> from PIL import Image as im
  >>>
  >>> s = np.array(im.open('./poglavje_10/proj2d_test.png'))
  >>> H, W = s.shape
  >>> x = np.arange(-2*(W-1)*0.5, 2*(W-1)*0.5 + 1, 2)
  >>> y = np.arange(-2*(H-1)*0.5, 2*(H-1)*0.5 + 1, 2)
  >>> fi = np.arange(0, 2*np.pi, np.pi/180)
  >>>
  >>> p = imFanBeamProject2d(fi, 2, s, x, y, 800, 512, 2.0)
  >>> p_f = filterSinogram(p, 'hamming')
  >>>
  >>> sr = imFanBeamBackproject2D(fi, p, x, y, 800, 2.0)
  >>> sr_f = imFanBeamBackproject2D(fi, p_f, x, y, 800, 2.0)
  >>>
  >>> pp.figure()
  >>> imShowEx(s, title='Testna slika', subplot=221, extent=[x[0], x[-1], y[0], y[-1]])
  >>> imShowEx(p, title='Sinogram', subplot=222)
  >>> imShowEx(sr, title='Rekonstrukcija brez filtriranjem', subplot=223)
  >>> imShowEx(sr_f, title='Rekonstrukcija s filtriranjem', subplot=224)
  >>>
  '''
  # priprava podatkov
  if isinstance(fi, float) or isinstance(fi, int): 
    fi = [fi]
  fi = np.asarray(fi, dtype=np.float)
  sinogram = np.asarray(sinogram, dtype=np.float)
  x = np.asarray(x, dtype=np.float)
  y = np.asarray(y, dtype=np.float)
  d = float(d)
  pixelSize = float(pixelSize)

  # število slikovnih elementov linijskega tipala
  n = sinogram.shape[1]

  # koordinate slikovnih elementov rekonstruirane slike
  [Xg, Yg] = np.meshgrid(x, y)
  # koordinate razporedimo v dvovrstično matriko - x v prvo, y v drugo vrstico
  pts = np.vstack([Xg.flatten(), Yg.flatten()])

  # akumulator povratne projekcije enake velikosti kot sta Xg in Yg
  oBP = np.zeros(Yg.shape)
  for i in range(fi.size):
    # rotacija slikovnih elementovb v obratni smeri 
    #  kot bi rotirali tipalo-izvor
    Tr = np.array([[np.cos(-fi[i]), -np.sin(-fi[i])],
                   [np.sin(-fi[i]),  np.cos(-fi[i])]])
    ptsr = np.dot(Tr, pts)
    # y koordinate projiciranih slikovnih elementov na tipalu izražene z
    #  naslovom tipala
    yt = 2.0*ptsr[1,:]/(1.0 - 2.0*ptsr[0,:]/d)/ \
        pixelSize + n*0.5
    # poiščemo projekcije z veljavnimi naslovi
    validinds = (yt >= 0) & (yt <= n - 1)
    yt = yt[validinds]

    # z linearno interpolacijo določimo vrednosti v projiciranih točkah
    ind = np.floor(yt)
    delta = yt - ind
    intind = ind.astype('int')
    vt = sinogram[i, intind]*(1.0 - delta) + \
        sinogram[i, np.minimum(intind + 1, n - 1)]*delta
    validinds.shape = Xg.shape

    # prištejemo izračunane vrednosti akumulatorju povratne projekcije
    oBP[validinds] += vt

  return oBP

if __name__ == '__main__':
  #%% uvozi potrebne module
  from matplotlib import pyplot as pp
  from mpl_toolkits.mplot3d import Axes3D
  import os

  try:
      os.mkdir('./tmp')
  except OSError:
      pass

  #%% imLoadRaw2d test
  slika = imLoadRaw2d('./vaja0/mr_217x181_uint8.raw', 217, 181)
  pp.figure()
  pp.imshow(slika, cmap='gray')
  pp.title('Slika velikosti {} in tipa {}'.format(
    str(slika.shape), str(slika.dtype)))  

  #%% imSaveRaw2d test
  slika = np.tile(np.arange(256, dtype=np.uint8), [128, 1]) # (128, 256)
  imSaveRaw2d('./tmp/gradient_256x128_uint8.raw', slika)
  slika = imLoadRaw2d('./tmp/gradient_256x128_uint8.raw', 256, 128)
  pp.figure()
  pp.imshow(slika, cmap='gray')
  pp.title('Slika velikosti {} in tipa {}'.format(
    str(slika.shape), str(slika.dtype)))  

  #%% imLoadRaw3d test
  slika = imLoadRaw3d('./vaja0/mr_217x181x181_uint8.raw', 217, 181, 181)
  pp.figure()
  pp.imshow(slika[90], cmap='gray')
  pp.title('Prerez [90, :, :] slike velikosti {} in tipa {}'.format(
    str(slika.shape), str(slika.dtype)))      

  #%% imSaveRaw3d test
  slika = imLoadRaw3d('./vaja0/mr_217x181x181_uint8.raw', 217, 181, 181)
  imSaveRaw3d('./tmp/mr_100x181x181_uint8.raw', slika[:, :, :100])
  slika = imLoadRaw3d('./tmp/mr_100x181x181_uint8.raw', 100, 181, 181)
  pp.imshow(slika[50], cmap='gray')
  pp.title('Prerez [50, :, :] slike velikosti {} in tipa {}'.format(
    str(slika.shape), str(slika.dtype)))      

  #%% interp1d test
  xp = np.linspace(0, np.pi, 10)
  fp = np.sin(xp)
  x = np.linspace(0, np.pi, 30)

  pp. figure()

  pp.subplot(1, 2, 1)
  f_nearest = interp1d(x, xp, fp, 'nearest')
  pp.title('Interpolacija z najbližjim sosedom')
  pp.plot(xp, fp, '-xg', label='Podatki')
  pp.plot(x, f_nearest, '-or', label='Interpolacija')
  pp.xlim([0.0, np.pi])
  pp.legend()

  pp.subplot(1, 2, 2)
  f_linear = interp1d(x, xp, fp, 'linear')
  pp.title('Linearna interpolacija')
  pp.plot(xp, fp, '-xg', label='Podatki')
  pp.plot(x, f_linear, '-or', label='Interpolacija')
  pp.xlim([0.0, np.pi])
  pp.legend()

  #%% interp2d test
  xp = np.linspace(-1.0, 1.0, 20)
  yp = np.linspace(-1.0, 1.0, 10)
  Yp, Xp = np.meshgrid(yp, xp, indexing='ij')
  fp = Xp**2 + Yp**2
  x = np.linspace(-1.0, 1.0, 60)
  y = np.linspace(-1.0, 1.0, 30)
  Y, X = np.meshgrid(y, x, indexing='ij')
  f_linear = interp2d(X, Y, xp, yp, fp)
  f_nearest = interp2d(X, Y, xp, yp, fp, 'nearest')
  
  pp.figure()
  
  pp.subplot(1, 3, 1)
  pp.title('Podatki')
  pp.imshow(fp, cmap='gray')
  
  pp.subplot(1, 3, 2)
  pp.imshow(f_nearest, cmap='gray')
  pp.title('Interpolacija z najbližjim sosedom')
  
  pp.subplot(1, 3, 3)
  pp.imshow(f_linear, cmap='gray')
  pp.title('Bilinearn interpolacija')

  #%% interp3d test
  slika = imLoadRaw3d('./vaja0/mr_217x181x181_uint8.raw', 217, 181, 181)
  xp = np.arange(0, 217, 1)
  yp = zp = np.arange(0, 181, 1)
  
  x = 100
  y = z = np.arange(0, 181, 0.5)
  Z, Y, X = np.meshgrid(z, y, x, indexing='ij')
  prerez_nearest = np.squeeze(interp3d(X, Y, Z, xp, yp, zp, slika, 'nearest'))
  prerez_linear = np.squeeze(interp3d(X, Y, Z, xp, yp, zp, slika))

  pp.figure()

  pp.subplot(1, 3, 1)
  pp.title('Podatki')
  pp.imshow(slika[:,:,100], cmap='gray')

  pp.subplot(1, 3, 2)
  pp.imshow(prerez_nearest, cmap='gray')
  pp.title('Interpolacija z najbližjim sosedom')

  pp.subplot(1, 3, 3)
  pp.imshow(prerez_linear, cmap='gray')
  pp.title('Trilinearn interpolacija')

  #%% imGrid2d test
  Y, X = imGrid2d(10, 5)
  Yc, Xc = imGrid2d(10, 5, xoffset='center', yoffset='center')

  pp.figure()

  pp.subplot(1, 3, 1)
  pp.title('Brez zamika')
  pp.plot(X, Y, 'xr')

  pp.subplot(1, 3, 2)
  pp.title('Koordinatno izhodišče v težišču')
  pp.plot(Xc, Yc, 'og')

  pp.subplot(1, 3, 3)
  pp.title('Združeno')
  pp.plot(X, Y, 'xr')
  pp.plot(Xc, Yc, 'og')

  #%% imGrid3d test
  Z, Y, X = imGrid3d(20, 10, 5)
  Zc, Yc, Xc = imGrid3d(20, 10, 5, xoffset='center', yoffset='center', zoffset='center')

  fig = pp.figure()

  ax = fig.add_subplot(131, projection='3d')
  ax.plot(X.flatten(), Y.flatten(), Z.flatten(), 'xr')
  ax.set_title('Brez zamika')

  pp.subplot(1, 3, 2)
  ax = fig.add_subplot(132, projection='3d')
  ax.plot(Xc.flatten(), Yc.flatten(), Zc.flatten(), 'og')
  ax.set_title('Koordinatno izhodišče v težišču')

  pp.subplot(1, 3, 3)
  ax = fig.add_subplot(133, projection='3d')
  ax.plot(X.flatten(), Y.flatten(), Z.flatten(), 'xr')
  ax.plot(Xc.flatten(), Yc.flatten(), Zc.flatten(), 'og')
  ax.set_title('Združeno')
  
  #%% conv2d test
  konv_jedro = np.ones([5, 5])
  konv_jedro /= konv_jedro.sum()

  slika = imLoadRaw2d('./vaja0/mr_217x181_uint8.raw', 217, 181)
  filtrirana = conv2d(slika, konv_jedro)

  pp.figure()

  pp.subplot(1, 2, 1)
  pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
  pp.title('Slika')

  pp.subplot(1, 2, 2)
  pp.imshow(filtrirana, cmap='gray', vmin=0, vmax=255)
  pp.title('Filtrirana slika')
  
  #%% imDecimate2d test
  slika = imLoadRaw2d('./vaja0/mr_217x181_uint8.raw', 217, 181)
  filtrirana = imDecimate2d(slika, konv_jedro)

  pp.figure()

  pp.subplot(1, 5, 1)
  pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
  pp.title('Slika')

  for i in range(4):
    slika = imDecimate2d(slika)
    pp.subplot(1, 5, i + 2)
    pp.imshow(slika, cmap='gray', vmin=0, vmax=255)
    pp.title('Decimirana slika k={}'.format(2**(i + 1)))
